import logging
import warnings
import re
import copy

from uuid import uuid4
from threading import local
from django import template
from django.http import HttpRequest
from django.template import RequestContext, TemplateSyntaxError
from django.template.loader_tags import BlockNode, BLOCK_CONTEXT_KEY
from django.utils.safestring import mark_safe, SafeString

from ..exceptions import ComponentError, ComponentNotFound
from ..component_register import resolve_component
from ..utils import remove_surrounding_quotes

logger = logging.getLogger(__name__)


class TetraTemplateTagException(Exception):
    pass


thread_local = local()
register = template.Library()


def get_nodes_by_type_deep(obj, node_type):
    thread_local.geting_nodes_by_type_deep = True
    node_list = obj.get_nodes_by_type(node_type)
    thread_local.geting_nodes_by_type_deep = False
    return node_list


@register.simple_tag(takes_context=True, name="tetra_scripts")
def scripts_placeholder_tag(context, include_alpine=False):
    placeholder = f"<!-- tetra scripts {uuid4()} -->"
    try:
        context.request.tetra_scripts_placeholder_string = placeholder.encode()
        context.request.tetra_scripts_placeholder_include_alpine = include_alpine
    except AttributeError:
        raise TetraTemplateTagException(
            '{% tetra_scripts %} tag requires "request" in the template context.'
        )
    return mark_safe(placeholder)


@register.simple_tag(takes_context=True, name="tetra_styles")
def styles_placeholder_tag(context):
    placeholder = f"<!-- tetra styles {uuid4()} -->"  #
    try:
        context.request.tetra_styles_placeholder_string = placeholder.encode()
    except AttributeError:
        raise TetraTemplateTagException(
            '{% tetra_styles %} tag requires "request" in the template context.'
        )
    return mark_safe(placeholder)


ALL_CONTEXT = object()


def token_attr(bit, parser):
    if "=" not in bit:
        return {}
    name, value = bit.split("=", 1)
    return {name: parser.compile_filter(value)}


@register.tag(name="component")
def do_component(parser, token):
    split_contents = token.split_contents()
    if len(split_contents) < 2:
        raise TemplateSyntaxError("Component tag requires a component name")
    component_name = split_contents[1]
    bits = split_contents[2:]

    component_name = component_name.strip("'\"")

    # If the tag ends with a / than it has no content, otherwise it does.
    has_content = True
    if (len(bits) > 0) and (bits[-1] == "/"):
        has_content = False
        bits = bits[:-1]

    # bits can be split into sections stating with one of:
    # 'args:', 'attrs:' or 'context:'
    # The fist section defaults to 'args:' if it  is unprefixed.
    bits_grouped = {
        "args:": [],
        "attrs:": [],
        "context:": [],
    }
    current_prefix = "args:"
    for bit in bits:
        if bit in bits_grouped:
            current_prefix = bit
        else:
            bits_grouped[current_prefix].append(bit)

    # Args bits:
    args = []
    kwargs = {}
    for bit in bits_grouped["args:"]:
        # First we try to extract a potential kwarg from the bit
        kwarg = template.base.token_kwargs([bit], parser)
        if kwarg:
            # The kwarg was successfully extracted
            param, value = kwarg.popitem()
            if param in kwargs:
                # The keyword argument has already been supplied once
                raise TemplateSyntaxError(
                    f"Component '{component_name}' received multiple values for "
                    f"keyword argument '{param}'"
                )
            else:
                # All good, record the keyword argument
                kwargs[str(param)] = value
        else:
            if kwargs:
                raise TemplateSyntaxError(
                    f"Component '{component_name}' received some positional "
                    "argument(s) after some keyword argument(s)"
                )
            else:
                # Record the positional argument
                args.append(parser.compile_filter(bit))

    # Attrs bits:
    attrs = {}
    for bit in bits_grouped["attrs:"]:
        kwarg = token_attr(bit, parser)
        if kwarg:
            attr, value = kwarg.popitem()
            attrs[str(attr)] = value
        else:
            raise TemplateSyntaxError(
                f"Component '{component_name}' attrs must be prefixed by a attr name."
            )

    # Context bits:
    if "__all__" in bits_grouped["context:"]:
        if len(bits_grouped["context:"]) > 1:
            raise TemplateSyntaxError(
                f"__all__ and multiple context arguments are mutually exclusive in "
                f"Component '{component_name}'."
            )
        context_args = ALL_CONTEXT
    elif "**context" in bits_grouped["context:"]:
        # TODO: remove in 1.0
        if len(bits_grouped["context:"]) > 1:
            raise TemplateSyntaxError(
                f"Component '{component_name}' has multiple context arguments as well a "
                "**context for all context."
            )
        context_args = ALL_CONTEXT
        warnings.warn(
            f"Component '{component_name}': 'context: **context' is deprecated and "
            f"will be removed in future versions of Tetra. Please use 'context: __all__' instead.",
            DeprecationWarning,
        )
    else:
        context_args = {}
        for bit in bits_grouped["context:"]:
            # First we try to extract a potential kwarg from the bit
            kwarg = template.base.token_kwargs([bit], parser)
            if kwarg:
                # The kwarg was successfully extracted
                param, value = kwarg.popitem()
                if param in context_args:
                    # The context argument has already been supplied once
                    raise TemplateSyntaxError(
                        f"Component '{component_name}' received multiple values for "
                        f"context argument '{param}'"
                    )
                else:
                    # All good, record the keyword argument
                    context_args[str(param)] = value
            else:
                # Record the unprefixed argument using the var name as the context arg.
                # so my_context_var is interpreted as my_context_var=my_context_var
                context_args[bit] = parser.compile_filter(bit)

    nodelist = None
    if has_content:
        # Parse the contents of the use tag, we reset the parser.__loaded_slots
        # so that we can reuse slot names inside each individual use tag
        current_loaded_slots = getattr(parser, "__loaded_slots", None)
        parser.__loaded_slots = []
        nodelist = parser.parse((f"/{component_name}",))
        if current_loaded_slots is not None:
            parser.__loaded_slots = (
                current_loaded_slots  # Return original __loaded_slots
            )
        parser.delete_first_token()

    return ComponentNode(
        component_name,
        args,
        kwargs,
        attrs=attrs,
        context_args=context_args,
        nodelist=nodelist,
        origin=parser.origin,
    )


class ComponentNode(template.Node):
    def __init__(
        self,
        component_name,
        args,
        kwargs,
        attrs=None,
        context_args=None,
        nodelist=None,
        origin=None,
    ):
        self.component_name = component_name
        self.args = args
        self.kwargs = kwargs
        self.attrs = attrs
        self.context_args = context_args
        self.nodelist = nodelist
        self.slots = None
        self.prepare_slots(origin=origin)

    def __repr__(self):
        return f"<ComponentNode: {self.component_name}>"

    def get_nodes_by_type(self, nodetype):
        # This stops slots leeking out of scope.
        if nodetype == BlockNode and not getattr(
            thread_local, "geting_nodes_by_type_deep", False
        ):
            return []
        else:
            return super().get_nodes_by_type(nodetype)

    def prepare_slots(self, origin=None):
        if not self.nodelist:
            return
        default_slot = False

        for node in self.nodelist:
            if isinstance(node, template.base.TextNode) and re.match(r"^\s*$", node.s):
                continue
            if not isinstance(node, BlockNode):
                default_slot = True
                break

        if default_slot:
            slot = BlockNode("default", self.nodelist)
            slot.origin = origin
            self.slots = {"default": slot}
            self.nodelist = template.base.NodeList([slot])
            self.nodelist.contains_nontext = True
        else:
            self.slots = dict(
                (n.name, n) for n in self.nodelist.get_nodes_by_type(BlockNode)
            )

    def render(self, context: RequestContext) -> SafeString:
        """
        :param context: The template context in which the component is being rendered.
            It must include the "request" attribute.
        :return: The rendered component as a tag string or directly rendered component
            as per the resolved state.
        """
        # when component starts with "=", assume it is a dynamic variable name
        is_dynamic = self.component_name.startswith("=")
        if is_dynamic:
            # Handle dotted paths for dynamic component names
            path = self.component_name.split(".")
            path[0] = path[0][1:]
            # traverse the context for the component name
            c = context
            for part in path:
                try:
                    c = c[part]
                except TypeError:
                    c = getattr(c, part, None)
                except KeyError:
                    c = None

                if c is None:
                    raise ComponentNotFound(
                        f"Unable to resolve dynamic component: '{self.component_name}'"
                    )
            component_class = c
            if isinstance(component_class, str):
                component_class = resolve_component(context, component_class)
        else:
            component_class = resolve_component(context, self.component_name)

        try:
            request: HttpRequest = context.request
        except AttributeError:
            raise ComponentError(
                'Tetra Components require "request" in the template context.'
            )

        resolved_args = [var.resolve(context) for var in self.args]
        resolved_kwargs = {k: v.resolve(context) for k, v in self.kwargs.items()}
        resolved_attrs = {k: v.resolve(context) for k, v in self.attrs.items()}

        # check extra context arguments
        extra_context = getattr(component_class, "_extra_context", [])
        if type(extra_context) is str:
            extra_context = [extra_context]

        if self.context_args == ALL_CONTEXT or "__all__" in extra_context:
            resolved_context = context
        else:
            resolved_context = template.Context()
            ctx = {}
            for k in extra_context:
                if k in context:
                    ctx[k] = context[k]
                else:
                    # this variable is nnot in context!
                    logger.warning(
                        f"Component {self} uses '{k}' in _extra_context, "
                        f"but it is not available in current context."
                    )
            if ctx:
                resolved_context.update(ctx)

        # add "tetra" helper object to the context
        resolved_context["tetra"] = {
            "current_url": (
                request.tetra.current_url
                if request.tetra
                else request.build_absolute_uri()
            ),
            "current_url_path": (
                request.tetra.current_url_path if request.tetra else request.path
            ),
            "current_url_full_path": (
                request.tetra.current_url_full_path
                if request.tetra
                else request.get_full_path()
            ),
        }

        if self.context_args != ALL_CONTEXT:
            # update context with the explicitly given params. This may not
            # happen if __all__ context is requested directly on the template tag.
            resolved_context.update(
                {k: v.resolve(context) for k, v in self.context_args.items()}
            )

        # add "slots" dict to context, to easily filter out if each slot is available
        if self.slots:
            resolved_context["slots"] = {}
            for slot in self.slots:
                resolved_context["slots"][slot] = True

        slots = copy.copy(self.slots)
        if slots and BLOCK_CONTEXT_KEY in context.render_context:
            old_slot_context = context.render_context[BLOCK_CONTEXT_KEY]
            for slot_name, slot in slots.items():
                expose_as = getattr(slot, "expose_as", None)
                if expose_as:
                    new_slot = old_slot_context.get_block(expose_as)
                    if new_slot:
                        slots[slot_name] = new_slot

        children_state = context.get("_loaded_children_state", None)
        if "key" not in resolved_kwargs:
            resolved_kwargs["key"] = component_class.full_component_name()
        if (
            children_state
            and (resolved_kwargs["key"] in children_state)
            and not is_dynamic
        ):
            component_state = children_state[resolved_kwargs["key"]]
            component = component_class.from_state(
                component_state,
                request,
                *resolved_args,
                **resolved_kwargs,
                _attrs=resolved_attrs,
                _context=resolved_context,
                _slots=slots,
            )
            if "children" in component_state:
                component._loaded_children_state = component_state["children"]
            return component.render()
        else:
            return component_class.as_tag(
                request,
                *resolved_args,
                **resolved_kwargs,
                _attrs=resolved_attrs,
                _context=resolved_context,
                _slots=slots,
            )


@register.tag(name="...")
def do_attr_tag(parser, token):
    split_contents = token.split_contents()
    if len(split_contents) < 2:
        raise TemplateSyntaxError("Attr tag requires at least one argument")
    bits = split_contents[1:]

    # Args bits:
    args = []
    for bit in bits:
        # First we try to extract a potential kwarg from the bit
        kwarg = token_attr(bit, parser)
        if kwarg:
            # The kwarg was successfully extracted
            param, value = kwarg.popitem()
            args.append({str(param): value})
        else:
            args.append(parser.compile_filter(bit))

    return AttrsNode(args)


class AttrsNode(template.Node):
    def __init__(self, args):
        self.args = args

    def arg_gen(self, context):
        for arg in self.args:
            if isinstance(arg, template.base.FilterExpression):
                resolved_arg = arg.resolve(context)
                for key, value in resolved_arg.items():
                    yield key, value
            else:
                for key, value in arg.items():
                    resolved_value = value.resolve(context)
                    yield key, resolved_value
        # add subscribed events as Alpine x-on attr

        c = self.origin.component
        if hasattr(c, "_event_subscriptions"):
            for event_name, method_name in c._event_subscriptions.items():
                yield f"x-on:{event_name}", f"{method_name}($event.detail)"

    def render(self, context):
        attrs = {}
        for key, value in self.arg_gen(context):
            if key == "class":
                # TODO: class from boolean var
                attrs.setdefault("class", set())
                attrs["class"].update(value.split())
            elif key == "style":
                # TODO: advanced style via dict or dotted path from vars
                attrs.setdefault("style", {})
                for rule in value.split(";"):
                    prop_key, prop_val = rule.split(":", 1)
                    attrs["style"][prop_key] = prop_val
            else:
                attrs[key] = value
        if "class" in attrs:
            attrs["class"] = " ".join(attrs["class"])
        if "style" in attrs:
            attrs["style"] = "; ".join(f"{k}: {v}" for k, v in attrs["style"].items())
        for key, value in list(attrs.items()):
            if value is False:
                del attrs[key]
        return " ".join(k if v is True else f'{k}="{v}"' for k, v in attrs.items())


@register.tag("slot")
def do_slot(parser, token):
    """
    Define a slot that can be overridden by child templates.
    Based on the native Django "block" tag but adds an extra attribute (expose_as) to the
    BlockNode indicating if it is possible to override in an extending template and
    under what name.
    Syntax to expose a slot under its own name as block
    {% slot slot_name expose %}
    Syntax to expose a slot under a different (block) name:
    {% slot slot_name expose as exposed_block_name %}
    """
    # token.split_contents() isn't useful here because this tag doesn't accept
    # variable as arguments.
    bits = token.contents.split()
    if len(bits) not in (2, 3, 5):
        raise TemplateSyntaxError("'%s' tag takes one, two or four arguments" % bits[0])
    if len(bits) > 2 and bits[2] != "expose":
        raise TemplateSyntaxError(
            "'%s' tag second argument can only be 'expose' if given" % bits[0]
        )
    if len(bits) == 5 and bits[3] != "as":
        raise TemplateSyntaxError(
            "'%s' tag third argument can only be 'as' if given" % bits[0]
        )

    slot_name = bits[1]

    if len(bits) == 5:
        expose_as = bits[4]
    elif len(bits) == 3:
        expose_as = slot_name
    else:
        expose_as = None

    # Keep track of the names of BlockNodes found in this template, so we can
    # check for duplication.
    try:
        if slot_name in parser.__loaded_slots:
            raise TemplateSyntaxError(
                "'%s' tag with name '%s' appears more than once" % (bits[0], slot_name)
            )
        parser.__loaded_slots.append(slot_name)
    except AttributeError:  # parser.__loaded_slots isn't a list yet
        parser.__loaded_slots = [slot_name]
    nodelist = parser.parse(("endslot",))

    # This check is kept for backwards-compatibility. See #3100.
    endslot = parser.next_token()
    acceptable_endblocks = ("endslot", "endslot %s" % slot_name)
    if endslot.contents not in acceptable_endblocks:
        parser.invalid_block_tag(endslot, "endslot", acceptable_endblocks)

    node = BlockNode(slot_name, nodelist)
    node.expose_as = expose_as
    node.origin = parser.origin  # Needed for persistent_id when pickeling
    return node


@register.tag(name="livevar")
def live_variable(parser, token) -> template.Node:
    """Renders a tag with dynamically filled/reactive HTML content from a
    Tetra/Alpine.js variable."""
    split_contents = token.split_contents()
    if len(split_contents) > 3:
        raise TemplateSyntaxError(
            "livevar tag requires maximum two arguments: the variable name, "
            "and optionally the HTML tag."
        )
    if len(split_contents) == 3:
        tag = split_contents[2]
        if "=" not in tag:
            raise TemplateSyntaxError(
                f"livevar 'tag' argument expects a variable assignment "
                f"(tag=<variable_name>), got '{tag}'"
            )
        _, tag = tag.split("=")
        tag = remove_surrounding_quotes(tag)
    else:
        tag = "span"

    var_name = split_contents[1]

    var_name = remove_surrounding_quotes(var_name)

    return LiveVariableNode(var_name, tag)


class LiveVariableNode(template.Node):
    def __init__(self, var_name: str, tag: str):
        self.var_name = var_name
        self.tag = tag

    def __repr__(self):
        return f"<LiveVariableNode: {self.var_name}>"

    def render(self, context):
        return mark_safe(
            f"<{self.tag} x-text='{self.var_name}' "
            f"x-show='{self.var_name}'></{self.tag}>"
        )


@register.filter(name="if")
def if_filter(value, condition):
    if condition:
        return value
    else:
        return ""


@register.filter(name="else")
def else_filter(value, else_value):
    if value:
        return value
    else:
        return else_value



import logging
import os
import shutil
import subprocess
import json
import warnings
from collections import defaultdict
from typing import Self

from django.apps import AppConfig, apps
from django.templatetags.static import static
from django.utils.functional import cached_property

from .conf import get_setting, get_esbuild_path
from .components.base import ComponentMetaClass, BasicComponent, Component
from .exceptions import LibraryError
from .utils import camel_case_to_underscore

logger = logging.getLogger(__name__)


class Library:
    # a dictionary to store all Library instances: [app_label][library][components]
    registry: defaultdict[str, dict[str, Self]] = defaultdict(dict)

    @staticmethod
    def __new__(
        cls, name: str = "", app: AppConfig | str = None, path: str = ""
    ) -> Self:
        """Returns a new instance of Library, or the existing instance, if a library
        with the same app/name already exists."""

        if not name or not app:
            raise ValueError("Library 'name' and 'app' parameters are required.")
        if type(app) is str:
            app = apps.get_app_config(app)
        if app.label not in cls.registry or name not in cls.registry[app.label]:
            instance = super().__new__(cls)
            cls.registry[app.label][name] = instance
        else:
            instance = cls.registry[app.label][name]
        return instance

    def __init__(self, name: str, app: AppConfig | str, path: str = ""):
        if type(app) is str:
            app = apps.get_app_config(app)

        # Initialize only if this is a new instance
        if not hasattr(self, "components"):
            self.components: dict[str, type[BasicComponent] | ComponentMetaClass] = {}
            self.name = name
            self.path = path
            self.app: AppConfig = app

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<Library: {self.display_name}>"

    @property
    def display_name(self):
        return f"{getattr(self, 'app').label}.{getattr(self, 'name')}"

    @property
    def js_filename(self):
        """Returns the filename of the compiled app/library-wide JavaScript file."""
        return f"{self.app.label}_{self.name}.js"

    @property
    def styles_filename(self):
        """Returns the filename of the compiled app/library-wide CSS file."""
        return f"{self.app.label}_{self.name}.css"

    @property
    def js_path(self):
        """Returns the path to the compiled app/library-wide JavaScript file."""
        return os.path.join(
            self.app.path,
            "static",
            self.app.label,
            "tetra",
            self.name,
            self.js_filename,
        )

    @property
    def styles_path(self):
        return os.path.join(
            self.app.path,
            "static",
            self.app.label,
            "tetra",
            self.name,
            self.styles_filename,
        )

    @cached_property
    def js_url(self):
        """Returns the static URL of the library's compiled JavaScript file."""
        with open(f"{self.js_path}.filename") as f:
            js_filename = f.read()
        return static(os.path.join(self.app.label, "tetra", self.name, js_filename))

    @cached_property
    def styles_url(self):
        """Returns the static URL of the library's compiled CSS file."""
        with open(f"{self.styles_path}.filename") as f:
            styles_filename = f.read()
        return static(os.path.join(self.app.label, "tetra", self.name, styles_filename))

    def register(
        self,
        component_cls: type[BasicComponent] | type[Component],
        name: str | None = None,
    ):
        if not name:
            name = component_cls.__name__
        underscore_name = camel_case_to_underscore(name)

        def dec(cls: type[BasicComponent]):
            if hasattr(cls, "_library") and cls._library:
                if cls._library is not self:
                    # If this is a subclass of an already registered component,
                    # we need to make sure we don't just inherit the library
                    # and name.
                    if (
                        cls.__name__ == component_cls.__name__
                        and cls.__module__ == component_cls.__module__
                    ):
                        # This IS the same class, check if it was registered to a different library
                        if cls._library is not self:
                            raise LibraryError(
                                f"Error registering component '{component_cls.__name__}' to "
                                f"library {self.display_name}, as it is "
                                f"already registered to library {cls._library.display_name}."
                            )
                    else:
                        # This is a subclass, we can register it to a new library
                        pass
                else:
                    logger.warning(
                        f"The app {self.app} is trying to register the "
                        f"'{component_cls.__name__}' component, which is already "
                        f"registered to library {cls._library}. "
                        f"Ignoring second registration."
                    )
                    return cls
            component_cls._library = self
            component_cls._name = underscore_name
            self.components[underscore_name] = component_cls
            return cls

        def component_tag_compile_function(parser, token):
            # Modify token to include the "component" prefix
            tag = token.contents.split()[0]
            if tag == "@":
                warnings.warn(
                    "Use 'component' instead of '@' for component tags.",
                    DeprecationWarning,
                )
            # TODO: check if component with this name exists!
            if tag != "component":
                token.contents = (
                    f"component {tag} {' '.join(token.contents.split()[1:])}"
                )
            return do_component(parser, token)

        if component_cls:
            # Dynamically register a template tag with ComponentName and
            # library.ComponentName

            # Set the name and other attributes
            component_tag_compile_function.__name__ = name
            from .templatetags.tetra import do_component, register as tetra_register

            # Register the tag with Django's template system
            tetra_register.tag(
                name=name, compile_function=component_tag_compile_function
            )
            tetra_register.tag(
                name=f"{self.name}.{name}",
                compile_function=component_tag_compile_function,
            )

            return dec(component_cls)
        else:
            return dec

    def __contains__(self, component_name: str) -> bool:
        """Check if the library contains the given component name."""
        return component_name in self.components

    def build(self, force=False):
        library_cache_path = os.path.join(
            self.app.path, get_setting("TETRA_FILE_CACHE_DIR_NAME"), self.name
        )
        file_out_path = os.path.join(
            self.app.path, "static", self.app.label, "tetra", self.name
        )
        meta_path = os.path.join(library_cache_path, "build_meta.json")

        # Get current state of source files
        current_state = {}
        for component_name, component_cls in self.components.items():
            py_filename, _, _ = component_cls.get_source_location()
            if py_filename and os.path.exists(py_filename):
                current_state[py_filename] = os.path.getmtime(py_filename)

            for ext in ["js", "css"]:
                ext_path = component_cls._get_component_file_path_with_extension(ext)
                if ext_path and os.path.exists(ext_path):
                    current_state[ext_path] = os.path.getmtime(ext_path)

        # Check if we can skip build
        if not force and os.path.exists(meta_path) and os.path.exists(file_out_path):
            try:
                with open(meta_path, "r") as f:
                    saved_state = json.load(f)
                if saved_state == current_state:
                    # Check if expected output files exist
                    js_filename_path = f"{self.js_path}.filename"
                    styles_filename_path = f"{self.styles_path}.filename"

                    has_js = any(c.has_script() for c in self.components.values())
                    has_styles = any(c.has_styles() for c in self.components.values())

                    js_ok = not has_js or (
                        os.path.exists(js_filename_path)
                        and os.path.exists(
                            os.path.join(
                                file_out_path, open(js_filename_path).read().strip()
                            )
                        )
                    )
                    styles_ok = not has_styles or (
                        os.path.exists(styles_filename_path)
                        and os.path.exists(
                            os.path.join(
                                file_out_path, open(styles_filename_path).read().strip()
                            )
                        )
                    )

                    if js_ok and styles_ok:
                        return
            except Exception as e:
                print(f"Incremental build check failed: {e}")

        print(f"# Building {self.display_name}")

        # Clear existing files to prevent old versions
        if os.path.exists(file_out_path):
            shutil.rmtree(file_out_path)
        os.makedirs(file_out_path)

        # Also clear from STATIC_ROOT if it exists
        static_root = get_setting("STATIC_ROOT")
        if static_root:
            static_root_path = os.path.join(
                static_root, self.app.label, "tetra", self.name
            )
            if os.path.exists(static_root_path):
                shutil.rmtree(static_root_path)

        if os.path.exists(library_cache_path):
            # We don't rmtree the whole cache path because it might contain build_meta.json
            for item in os.listdir(library_cache_path):
                if item == "build_meta.json":
                    continue
                item_path = os.path.join(library_cache_path, item)
                if os.path.isdir(item_path):
                    shutil.rmtree(item_path)
                else:
                    os.remove(item_path)
        else:
            os.makedirs(library_cache_path)

        self.build_js(library_cache_path, file_out_path)
        self.build_styles(library_cache_path, file_out_path)

        # Save current state
        with open(meta_path, "w") as f:
            json.dump(current_state, f)

    def build_js(self, library_cache_path, target_path):
        main_imports = []
        main_scripts = []
        # files_to_remove = []
        out_file_path = os.path.join(library_cache_path, self.js_filename)
        meta_filename = f"{self.js_filename}__meta.json"
        meta_file_path = os.path.join(library_cache_path, meta_filename)

        try:
            for component_name, component_cls in self.components.items():
                print(f" - {component_name}")
                if component_cls.has_script():
                    script = component_cls.extract_script()
                    py_filename, _, _ = component_cls.get_source_location()
                    py_dir = os.path.dirname(py_filename)
                    if component_cls._is_script_inline():
                        filename = os.path.join(
                            library_cache_path,
                            f"{os.path.basename(py_filename)}__{component_name}.tmp.js",
                        )
                        with open(filename, "w") as f:
                            f.write(script)
                        # files_to_remove.append(filename)
                    else:
                        filename = os.path.join(py_dir, f"{component_name}.js")
                    rel_path = os.path.relpath(filename, library_cache_path)
                    if os.name == "nt":
                        rel_path = rel_path.replace(os.sep, "/")
                    if not rel_path.startswith("./") and not rel_path.startswith("../"):
                        rel_path = "./" + rel_path
                    main_imports.append(f'import {component_name} from "{rel_path}";')
                    main_scripts.append(component_cls.render_script(component_name))
                else:
                    main_scripts.append(component_cls.render_script())

            with open(out_file_path, "w") as f:
                f.write("\n".join(main_imports))
                f.write("\n\n")
                f.write("\n".join(main_scripts))

            esbuild_ret = subprocess.run(
                [get_esbuild_path(), out_file_path]
                + get_setting("TETRA_ESBUILD_JS_ARGS")
                + [f"--outdir={target_path}", f"--metafile={meta_file_path}"]
            )

            if esbuild_ret.returncode != 0:
                print("ERROR BUILDING JS:", self.display_name)
                return
        finally:
            # for path in files_to_remove:
            #     os.remove(path)
            pass

        with open(meta_file_path) as f:
            meta = json.load(f)
        for path, data in meta["outputs"].items():
            if data.get("entryPoint", None):
                out_path = path
                break

        with open(f"{self.js_path}.filename", "w") as f:
            f.write(os.path.basename(out_path))

    def build_styles(self, library_cache_path, target_path):
        main_imports = []
        files_to_remove = []
        out_file_path = os.path.join(library_cache_path, self.styles_filename)
        meta_filename = f"{self.styles_filename}__meta.json"
        meta_file_path = os.path.join(library_cache_path, meta_filename)

        try:
            for component_name, component_cls in self.components.items():
                if component_cls.has_styles():
                    print(f" - {component_name}")
                    styles = component_cls.extract_styles()
                    py_filename, _, _ = component_cls.get_source_location()
                    py_dir = os.path.dirname(py_filename)
                    if component_cls._is_styles_inline():
                        filename = os.path.join(
                            library_cache_path,
                            f"{os.path.basename(py_filename)}__"
                            f"{component_name}.tmp.css",
                        )
                        with open(filename, "w") as f:
                            f.write(styles)
                        # files_to_remove.append(filename)
                    else:
                        filename = os.path.join(py_dir, f"{component_name}.css")
                    rel_path = os.path.relpath(filename, library_cache_path)
                    if os.name == "nt":
                        rel_path = rel_path.replace(os.sep, "/")
                    main_imports.append(f"@import '{rel_path}';")

            with open(out_file_path, "w") as f:
                f.write("\n".join(main_imports))

            esbuild_ret = subprocess.run(
                [get_esbuild_path(), out_file_path]
                + get_setting("TETRA_ESBUILD_CSS_ARGS")
                + [
                    f"--outdir={target_path}",
                    # These three lines below are a work around so that urls to images
                    # update correctly.
                    "--metafile=meta.json",
                    f"--outbase={self.app.path}",
                    f"--asset-names={os.path.relpath(self.app.path, target_path)}/[dir]/[name]",
                    "--allow-overwrite",
                    f"--metafile={meta_file_path}",
                ]
            )

            if esbuild_ret.returncode != 0:
                print("ERROR BUILDING CSS:", self.display_name)
                return
        finally:
            for path in files_to_remove:
                os.remove(path)

        with open(meta_file_path) as f:
            meta = json.load(f)
        for path, data in meta["outputs"].items():
            if data.get("entryPoint", None):
                out_path = path
                break

        with open(f"{self.styles_path}.filename", "w") as f:
            f.write(os.path.basename(out_path))



from django.urls import path

from .views import render_component_view

urlpatterns = [
    path(
        "render_component_view",
        render_component_view,
        name="render_component_view",
    ),
]



"""Full stack component framework for Django using Alpine.js"""

from .components import BasicComponent, Component, public
from .library import Library

__all__: list[str] = [
    "BasicComponent",
    "Component",
    "public",
    "Library",
]

try:
    import channels  # noqa

    __all__.append("ReactiveComponent")
except ImportError:
    pass


def __getattr__(name):
    if name == "ReactiveComponent":
        try:
            import channels  # noqa
        except ImportError:
            raise ImportError(
                "ReactiveComponent requires 'channels' to be installed. "
                "Please install it with: pip install channels"
            ) from None
        from .components.reactive import ReactiveComponent

        return ReactiveComponent
    raise AttributeError(f"module {__name__} has no attribute {name}")


__version__ = "0.8.0"
__version_info__ = tuple([num for num in __version__.split(".")])



from django.core.management.base import BaseCommand
from tetra.utils import cleanup_temp_uploads
from django.utils.translation import gettext as _


class Command(BaseCommand):
    help = _("Clean up old Tetra temporary upload files (default: older than 24 hrs.")

    def add_arguments(self, parser):
        parser.add_argument(
            "--max-age-hours",
            type=int,
            default=24,
            help=_("Maximum age of temporary files in hours (default: 24)"),
        )

    def handle(self, *args, **options):
        count = cleanup_temp_uploads(options["max_age_hours"])
        if count == 0:
            self.stdout.write(
                self.style.SUCCESS(_("No temporary upload files found to clean up."))
            )
            return

        self.stdout.write(
            self.style.SUCCESS(
                _("Successfully cleaned up {count} temporary upload files.").format(
                    count=count
                )
            )
        )



import logging

from typing import Set
from django.contrib.auth.models import AnonymousUser, AbstractUser
from channels.generic.websocket import AsyncJsonWebsocketConsumer
from django.contrib.sessions.backends.file import SessionStore

from tetra.components import subscription
from tetra.exceptions import ComponentError

logger = logging.getLogger(__name__)


class TetraConsumer(AsyncJsonWebsocketConsumer):
    """
    A consumer that handles WebSocket connections for Tetra.

    The data is sent as JSON to the client in the following format:

        {
          "type": "subscribe|unsubscribe|notify|component_update|component_remove",
          "channel": "channel_name",
          "component_id": "component_id",
          "event_name": "event_name",
          "data": {},
          "sender_id": "sender_id"  # optional
          "status": 200|400,
        }
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.user: AbstractUser | None = None
        self.session: SessionStore | None = None
        self.component_id: str | None = None
        self.subscribed_groups: Set[str] = set()

    async def connect(self) -> None:
        """
        Establish WebSocket connection and auto-subscribe to user, session, and broadcast groups.

        Closes connection if no session is available.
        """
        # Connect to session-specific channel
        self.session = self.scope.get("session")
        if not self.session:
            # Handle case where session is not available
            # There is no channels connection possible then.
            await self.close()
            return

        await self.accept()

        # Auto subscriptions: user, session, broadcast

        # subscribe client to user-specific group if authenticated
        self.user = self.scope.get("user", AnonymousUser())
        if self.user.is_authenticated:
            await self.channel_layer.group_add(
                f"user.{self.user.id}", self.channel_name
            )
            self.subscribed_groups.add(f"user.{self.user.id}")
            logger.debug(
                f"Subscribed '{self.channel_name}' to 'user.{self.user.id}' group."
            )

        # session keys theoretically could be None in an invalid session.
        if self.session.session_key:
            # subscribe client to session-specific group
            await self.channel_layer.group_add(
                f"session.{self.session.session_key}", self.channel_name
            )
            self.subscribed_groups.add(f"session.{self.session.session_key}")
            logger.debug(
                f"Subscribed '{self.channel_name}' to 'session.{self.session.session_key}' "
                f"group."
            )

        # Connect to public broadcast group
        await self.channel_layer.group_add("broadcast", self.channel_name)
        self.subscribed_groups.add("broadcast")
        logger.debug(f"Subscribed '{self.channel_name}' to 'broadcast' group.")

    async def disconnect(self, code) -> None:
        """Disconnects from all subscribed groups."""
        for group in self.subscribed_groups:
            await self.channel_layer.group_discard(group, self.channel_name)
            logger.debug(f"Discarded '{self.channel_name}' from '{group}' group.")
        # remove all subscribed groups after disconnect
        self.subscribed_groups.clear()

    async def receive_json(self, content, **kwargs):
        """Handle incoming JSON messages.

        This method processes different types of messages based on the 'type' field
        in the received JSON content and routes them to appropriate handler methods.
        Supported message types include 'subscribe', 'unsubscribe', 'notify', and
        'data_request'. Unknown message types are logged as warnings.
        """
        message_type = content.get("type")
        if message_type == "subscribe":
            await self._handle_subscribe(content)
        elif message_type == "unsubscribe":
            await self._handle_unsubscribe(content)
        else:
            logger.warning(
                f"Unknown message type received via websocket: {message_type}"
            )

    async def _handle_subscribe(self, data: dict[str, str]) -> None:
        """Handles subscription requests to named groups."""
        group_name = data.get("group", "")
        if not group_name:
            await self.send_json(
                {
                    "type": "subscription.response",
                    "group": group_name,
                    "status": "error",
                    "message": "Invalid group name",
                }
            )
            return

        # TODO: check if group name is valid, in a registry

        if group_name.startswith("user."):
            user_id = group_name.split(".")[1]
            # TODO: admin group instead of is_superuser
            if (
                self.user
                and str(self.user.id) != user_id
                and not self.user.is_superuser
            ):
                logger.warning(
                    f"Unauthorized access to group {group_name} by user "
                    f"{user_id} blocked."
                )
                await self.send_json(
                    {
                        "type": "subscription.response",
                        "group": group_name,
                        "status": "error",
                        "message": "Unauthorized",
                    }
                )
                return
        if group_name.startswith("session."):
            # you cannot manually subscribe to a session group.
            # While we could check against the current valid session key, it is way
            # too dangerous, enabling XSS attacks. So deny that completely.
            logger.warning(
                f"Unauthorized access to group {group_name} by session "
                f"{self.session.session_key if self.session else None} blocked."
            )
            await self.send_json(
                {
                    "type": "subscription.response",
                    "group": group_name,
                    "status": "error",
                    "message": "No manually joining of session groups allowed.",
                }
            )
        # check if we already subscribed to this group - if so, send "resubscribed"
        if group_name in self.subscribed_groups:
            await self.send_json(
                {
                    "type": "subscription.response",
                    "group": group_name,
                    "status": "resubscribed",
                }
            )
            return

        # Join channel group
        await self.channel_layer.group_add(group_name, self.channel_name)
        self.subscribed_groups.add(group_name)
        await self.send_json(
            {
                "type": "subscription.response",
                "group": group_name,
                "status": "subscribed",
            }
        )

    async def _handle_unsubscribe(self, data: dict[str, str]) -> None:
        group_name = data.get("group")

        if group_name in self.subscribed_groups:
            await self.channel_layer.group_discard(group_name, self.channel_name)
            self.subscribed_groups.remove(group_name)
            await self.send_json(
                {
                    "type": "subscription.response",
                    "group": group_name,
                    "status": "unsubscribed",
                }
            )

    async def subscription_response(self, event) -> None:
        """Handle confirmation of subscription"""
        await self.send_json(event)

    async def component_update_data(self, event) -> None:
        """Handle component data update

        This method is automatically called when a message is sent via group_send
        with the type "component.update_data". It checks if the component has matching fields
        and if so, sends the message to the client via websocket.
        """
        group_name = event["group"]
        components = subscription.registry.get(group_name, [])
        for c in components:
            if all(key in c._public_properties for key in event["data"].keys()):
                await self.send_json(event)
            else:
                raise ComponentError(
                    f"No matching components found for "
                    f"channel group data {event['data'].keys()} in '{group_name}'."
                )
        if not components:
            logger.warning(
                f"No matching component found for channel group {group_name}."
            )

    async def component_remove(self, event) -> None:
        """Handle component removal"""
        await self.send_json(event)

    async def notify(self, event) -> None:
        """Handle notification"""
        await self.send_json(event)



from django.urls import path
from . import views

urlpatterns = [
    path(
        "<str:app_name>/<str:library_name>/<str:component_name>/<str:method_name>",
        views.component_method,
        name="tetra_public_component_method",
    ),
]



class CallbackPath:
    def __init__(self, root, path=("",)):
        self.root = root
        self.path = path

    def __getattr__(self, name):
        return CallbackPath(self.root, self.path + (name,))

    def __getitem__(self, name):
        return self.__getattr__(name)

    def __call__(self, *args):
        self.root.callbacks.append(
            {
                "callback": self.path,
                "args": args,
            }
        )


class CallbackList:
    def __init__(self):
        self.callbacks = []

    def __getattr__(self, name):
        return CallbackPath(self, (name,))

    def __getitem__(self, name):
        return self.__getattr__(name)

    def serialize(self):
        return self.callbacks



import logging
import os
import pkgutil

from django.apps import apps
from django.conf import settings
import importlib
from django.template import Template
import inspect
from collections import defaultdict

from .components.base import InlineTemplate, Component, BasicComponent
from .exceptions import ComponentError, ComponentNotFound
from .library import Library
from .utils import (
    camel_case_to_underscore,
    underscore_to_pascal_case,
    unsupported_modules,
    is_abstract,
)
from .constants import COMPONENT_MODULE_NAMES

logger = logging.getLogger(__name__)


def find_component_libraries():
    """Finds libraries in component modules of all installed django apps."""
    from . import state

    if state.find_libraries_done:
        return
    state.loading_libraries = True
    importlib.invalidate_caches()
    for components_module_name in getattr(
        settings, "TETRA_COMPONENTS_MODULE_NAMES", COMPONENT_MODULE_NAMES
    ):
        for app_config in [
            a
            for a in apps.get_app_configs()
            if a.module.__name__ not in unsupported_modules
        ]:
            # from django.utils.module_loading import *

            module_name = f"{app_config.name}.{components_module_name}"
            try:
                components_module = importlib.import_module(module_name)

                # iter over a list of submodules in the
                # components directory
                for module_finder, library_name, ispkg in pkgutil.iter_modules(
                    components_module.__path__
                ):
                    # if library name is already registered, get the instance, else register it.
                    library = Library(name=library_name, app=app_config)

                    # if submodule is a package, treat it as library package
                    library_module_name = ".".join(
                        [app_config.name, components_module_name, library_name]
                    )
                    try:
                        library_module = importlib.import_module(library_module_name)
                    except ModuleNotFoundError as e:
                        # logger.critical(e)
                        raise ModuleNotFoundError(
                            f"Could not import library module '{library_module_name}': {e}"
                        )

                    library.path = os.path.join(
                        module_finder.path,
                        library_name,
                    )

                    # Search for components directly defined in the library package.
                    # this could be in e.g. default.py or in default/__init__.py
                    for name, member in inspect.getmembers(
                        library_module, inspect.isclass
                    ):
                        # accept only BasicComponent subclasses, and ignore imports:
                        # only classes defined in that module are registered as
                        # components.
                        if (
                            issubclass(member, BasicComponent)
                            and getattr(member, "__module__", None)
                            == library_module_name
                            and not is_abstract(member)
                        ):
                            member._is_directory_component = False
                            library.register(member, name)

                    # if library is a package, search for component packages within
                    # library
                    if ispkg:
                        for component_info in pkgutil.iter_modules([library.path]):
                            components_found = 0
                            component_name = ".".join(
                                [library_module_name, component_info.name]
                            )
                            try:
                                component_module = importlib.import_module(
                                    component_name
                                )
                            except ModuleNotFoundError as e:
                                # logger.critical(e)
                                raise ModuleNotFoundError(
                                    f"Could not import module '{component_name}': {e}"
                                )

                            for name, member in inspect.getmembers(
                                component_module, inspect.isclass
                            ):
                                # accept only BasicComponent subclasses, and ignore imports:
                                # only classes defined in that module are registered as
                                # components.
                                if (
                                    issubclass(member, BasicComponent)
                                    and getattr(member, "__module__", None)
                                    == component_module.__name__
                                    and not is_abstract(member)
                                ):
                                    components_found += 1
                                    if components_found > 1:
                                        raise ComponentError(
                                            f"Multiple components found "
                                            f"in '{component_module.__name__}' in app '{app_config.label}'."
                                            f"This is not supported"
                                        )
                                    member._is_directory_component = True
                                    library.register(member, name)

                    # else:
                    #     # TODO: library is a file, register all components in it
                    #     #  automatically.
                    #     for name, member in inspect.getmembers(components_module):
                    #         if isinstance(member, Library):
                    #             if name in libraries[app_config.label]:
                    #                 raise ComponentLibraryException(
                    #                     f'Library named "{name}" already in app "{app_config.label}".'
                    #                 )
                    #             libraries[app_config.label][name] = member
                    #             member.name = name
                    #             member.app = app_config

            except ModuleNotFoundError as e:
                # only raise the exception if the import raises secondary import errors.
                # E.g. if the component imports a module that is non-existent.
                if e.name != module_name:
                    raise e

                # if the module just is not present, just ignore it - this just means
                # that this app does not have a "components" package, which is ok.

    state.find_libraries_done = True
    state.loading_libraries = False

    BasicComponent._compile_all_templates()


def resolve_component(context, name: str) -> Component:
    """Takes a Django context and a component name, and returns the corresponding
    component instance.

    Attributes:
        context (RequestContext): The Django context in which the component is being rendered.
        name (str): The name of the component to resolve. This can be a full component name,
            or a dot-separated string representing a context variable containing a
            component.
        Returns:
            The resolved component instance.
    """
    dynamic = False
    name_parts = name.split(".")
    error_message = ""
    if name.startswith("="):
        dynamic = True
        name_parts[0] = name_parts[0][1:]

    if dynamic:
        # traverse context
        value = context
        for key in name_parts:
            try:
                # try to get component as key from context dict
                value = value[key]
            except KeyError:
                raise ComponentNotFound(
                    f"Dynamic component '{name}' not found in context."
                )
            except TypeError:
                # then this name part is already an object.
                # try to read attribute from object
                value = getattr(value, key, None)
        if value is not None:
            logger.debug(f"Resolved dynamic component: '{name}' to {Component}")
            return value
    else:
        try:
            if len(name_parts) > 2:
                raise ComponentNotFound(
                    f'Component name "{name}" invalid, should be in form '
                    '"[library_name.]ComponentName".'
                )
            if len(name_parts) == 1:
                name_parts = ("default", name_parts[0])

            # Full component name, easy: <library>.<component>!
            for app_name in Library.registry:
                try:
                    if name_parts[0] in Library.registry[app_name]:
                        # return the first matching library that contains the component
                        return Library.registry[app_name][name_parts[0]].components[
                            camel_case_to_underscore(name_parts[1])
                        ]
                except KeyError:
                    continue
            else:
                raise ComponentNotFound(
                    f"Component '{name_parts[1]}' not found in library '{name_parts[0]}'."
                )

        except ComponentNotFound as e:
            error_message = str(e)

    # if no method lead to finding a component successfully, give the user a hint
    # which components are available.
    components = []
    for app_name, lib in Library.registry.items():
        if lib:
            for lib_name, library in lib.items():
                if library.components:
                    for component_name in library.components:
                        ln = f"{lib_name}." if lib_name != "default" else ""
                        components.append(
                            f"{ln}{underscore_to_pascal_case(component_name)}"
                        )
    if not error_message:
        error_message = f'Component "{name}" not found.'
    error_message += f"\nAvailable components are: {components}"

    raise ComponentNotFound(error_message)



import sys

from django.conf import settings
from django.core.management import CommandError
from django.core.management.commands.runserver import Command as BaseRunserverCommand
from tetra.build import runserver_build


class Command(BaseRunserverCommand):
    # def inner_run(self, *args, **options):
    #     runserver_build()
    #     super().inner_run(*args, **options)

    def handle(self, *args, **options):
        runserver_build()
        # instead of super().handle(*args, **options)
        self.call_next_runserver(*args, **options)

    def call_next_runserver(self, *args, **options):
        from django.core.management import load_command_class

        # Get list of all apps excluding 'tetra*'
        apps = [app for app in settings.INSTALLED_APPS if not app.startswith("tetra")]
        # add 'django.core' as fallback, if even staticfiles is not there
        apps.append("django.core")

        for app_name in apps:
            try:
                # Attempt to load runserver from next app
                cmd = load_command_class(app_name, "runserver")
                cmd.run_from_argv(sys.argv)
                return
            except ModuleNotFoundError:
                continue
            except AttributeError:
                continue

        raise CommandError("No 'runserver' command found in any of the INSTALLED_APPS.")



from .base import (
    BasicComponent,
    Component,
    FormComponent,
    public,
    ModelFormComponent,
    DynamicFormMixin,
)

from ..exceptions import ComponentError, ComponentNotFound

__all__ = [
    ComponentError,
    ComponentNotFound,
    BasicComponent,
    Component,
    ModelFormComponent,
    public,
    DynamicFormMixin,
    FormComponent,
    # Router,
    # Link,
    # Redirect,
]



from glob import glob
from django.apps import AppConfig
from pathlib import Path
import os

from . import Library
from .templates import monkey_patch_template
from django.utils.autoreload import autoreload_started

monkey_patch_template()


def watch_extra_files(sender, *args, **kwargs):
    watch = sender.extra_files.add
    for app_name, library in Library.registry.items():
        for lib_name, library_info in library.items():
            if library_info and library_info.path:
                # watch for html, js, and css files
                watch_list = glob(f"{library_info.path}/**/*.*", recursive=True)
                for file in watch_list:
                    if os.path.exists(file) and file.endswith((".html", ".css", ".js")):
                        watch(Path(file))


class TetraConfig(AppConfig):
    name = "tetra"
    has_reactive_components = False

    def ready(self):
        from .component_register import find_component_libraries
        from tetra import checks  # noqa
        from .library import Library
        from tetra.router import Redirect
        from tetra.router import Link
        from tetra.router import Router

        # Register router components in the default library
        default_lib = Library("default", "tetra")
        default_lib.register(Router)
        default_lib.register(Link)
        default_lib.register(Redirect)

        find_component_libraries()
        autoreload_started.connect(watch_extra_files)



from django.apps import apps
from django.core.checks import Error, Warning, register, Tags
from django.conf import settings


@register(Tags.compatibility)
def check_websocket_support(app_configs, **kwargs):
    """
    Check if WebSocket support is properly configured for Tetra reactive components.

    The checks are only executed if there is demand for websockets - if any app
    declares a ReactiveComponent.
    """
    errors = []

    # Only check if reactive components are being used
    if not apps.get_app_config("tetra").has_reactive_components:
        return errors

    # Check if Channels is installed
    try:
        import channels
    except ImportError:
        errors.append(
            Error(
                "Django Channels is required for Tetra reactive components.",
                hint="Install channels: pip install channels",
                id="tetra.E001",
            )
        )
        return errors

    # Check ASGI_APPLICATION setting
    if not hasattr(settings, "ASGI_APPLICATION"):
        errors.append(
            Error(
                "ASGI_APPLICATION setting is required for WebSocket support.",
                hint="Add ASGI_APPLICATION = 'your_project.asgi.application' to settings.py",
                id="tetra.E002",
            )
        )
        return errors

    # Check if ASGI application exists and supports WebSockets
    asgi_app_path = getattr(settings, "ASGI_APPLICATION", None)
    if not asgi_app_path:
        errors.append(
            Error(
                "ASGI_APPLICATION setting is empty.",
                id="tetra.E003",
            )
        )
        return errors

    try:
        from django.utils.module_loading import import_string

        asgi_app = import_string(asgi_app_path)

        if not _has_websocket_support(asgi_app):
            errors.append(
                Warning(
                    "ASGI application does not appear to support WebSockets.",
                    hint="Ensure your ASGI application includes WebSocket routing for Tetra.",
                    id="tetra.W001",
                )
            )
    except (ImportError, AttributeError) as e:
        errors.append(
            Error(
                f"Cannot import ASGI application: {e}",
                hint=f"Check that '{asgi_app_path}' is correct and importable.",
                id="tetra.E004",
            )
        )

    # Check channel layer configuration
    try:
        from channels.layers import get_channel_layer

        layer = get_channel_layer()
        if layer is None:
            errors.append(
                Error(
                    "No channel layer configured.",
                    hint="Configure CHANNEL_LAYERS in settings.py for WebSocket support.",
                    id="tetra.E005",
                )
            )
    except Exception as e:
        errors.append(
            Warning(
                f"Channel layer configuration issue: {e}",
                id="tetra.W003",
            )
        )

    return errors


def _has_websocket_support(asgi_app) -> bool:
    """Check if an ASGI application supports WebSockets."""
    from channels.routing import ProtocolTypeRouter, URLRouter

    # Direct ProtocolTypeRouter check
    if isinstance(asgi_app, ProtocolTypeRouter):
        return "websocket" in asgi_app.application_mapping

    # Check for wrapped applications
    if hasattr(asgi_app, "application") or hasattr(asgi_app, "app"):
        wrapped_app = getattr(asgi_app, "application", None) or getattr(
            asgi_app, "app", None
        )
        if wrapped_app and wrapped_app != asgi_app:
            return _has_websocket_support(wrapped_app)

    return False



from asgiref.sync import async_to_sync
from django.apps import apps

from .base import Component, public
from .subscription import registry
from ..dispatcher import ComponentDispatcher


# Global flag to track if reactive components are registered

# a registry to keep track of all reactive components instances


class ReactiveComponent(Component):
    """A component that automatically connects to a channel and receives push
    notifications."""

    __abstract__ = True
    group_subscription: str = ""

    def __init_subclass__(cls, **kwargs):
        """Initialize the reactive component."""
        super().__init_subclass__(**kwargs)

        # set a global flag to indicate that reactive components are in use
        apps.get_app_config("tetra").has_reactive_components = True

    def __init__(self, *args, **kwargs):
        if "subscribe" in kwargs:
            if self.group_subscription:
                raise ValueError(
                    "Subscriptions cannot be defined in both component and template."
                )
            self.group_subscription = kwargs["subscribe"].strip()
        super().__init__(*args, **kwargs)

        # register the component with the channel layer
        # this is a list that can be queried for all reactive components to determine
        # their public_properties when sending data over websockets.
        group_name = self.group_subscription
        if not self.group_subscription in registry:
            registry[group_name] = [self]
        else:
            registry[group_name].append(self)

    def get_extra_tags(self) -> dict[str, str | None]:
        extra_tags = super().get_extra_tags()
        extra_tags["tetra-reactive"] = ""

        # Add subscribe channel groups as data attribute
        if self.group_subscription:
            extra_tags["tetra-subscription"] = self.group_subscription

        return extra_tags

    def get_subscription(self) -> str:
        """Returns the subscribed group for this component.

        This can be overridden to dynamically subscribe to channels when the
        component is created. Per default, it is evaluated once at initialization,
        not with every component cycle's load().
        """
        return self.group_subscription

    @public(update=False)
    def remove_component(self):
        async_to_sync(ComponentDispatcher.component_remove)(
            self.get_subscription(), self.component_id
        )



# Default settings for Tetra

TETRA_FILE_CACHE_DIR_NAME = "__tetracache__"
TETRA_ESBUILD_JS_ARGS = [
    "--bundle",
    "--minify",
    "--sourcemap",
    "--entry-names=[name]-[hash]",
    "--target=chrome80,firefox73,safari13,edge80",
]
TETRA_ESBUILD_CSS_ARGS = [
    "--bundle",
    "--minify",
    "--sourcemap",
    "--entry-names=[name]-[hash]",
    "--loader:.png=file",
    "--loader:.svg=file",
    "--loader:.gif=file",
    "--loader:.jpg=file",
    "--loader:.jpeg=file",
    "--loader:.webm=file",
    "--loader:.woff=file",
    "--loader:.woff2=file",
    "--loader:.ttf=file",
]

TETRA_TEMP_UPLOAD_PATH = "tetra_temp_upload"



from pathlib import Path
from django.apps import apps
from django.template.loaders.filesystem import Loader as FileSystemLoader
from django.conf import settings
from tetra.constants import COMPONENT_MODULE_NAMES


class Loader(FileSystemLoader):
    """Loader that loads templates from "components" directories in INSTALLED_APPS packages."""

    # @functools.lru_cache
    def get_dirs(self):
        # Collect all `components` directories from each app
        component_dirs = []
        for app_config in apps.get_app_configs():
            if app_config.label != "tetra":
                # TODO use dynamic components_module_names
                components_dir = Path(app_config.path) / getattr(
                    settings, "TETRA_COMPONENTS_MODULE_NAMES", COMPONENT_MODULE_NAMES
                )
                if components_dir.is_dir():
                    component_dirs.append(components_dir)
        return component_dirs



from channels.testing import WebsocketCommunicator
from django.apps import apps
from django.http import HttpRequest
from django.test import RequestFactory
from typing import AsyncGenerator, Any

import pytest
from django.urls import reverse
from django.contrib.sessions.backends.cache import SessionStore
from playwright.sync_api import Page

from tetra import Library
from tetra.components.base import BasicComponent
from tetra.middleware import TetraDetails

# These pytest fixtures may be usable in apps that use Tetra, so they are provided in a
# separate `fixtures` module that can be imported in your conftest.py


def add_session_to_request(request: HttpRequest) -> HttpRequest:
    """
    Helper to add a session to a factory request
    """
    from django.contrib.auth.models import AnonymousUser

    request.session = SessionStore()
    request.session.create()
    request.user = AnonymousUser()
    return request


@pytest.fixture
def post_request_with_session():
    """
    Returns an Http GET Request with a session, and a `tetra` attribute.
    """
    factory = RequestFactory()
    request = factory.post("/")  # Create a request object

    add_session_to_request(request)

    return request


@pytest.fixture
def current_app():
    return apps.get_app_config("main")


@pytest.fixture
def component_locator(page: Page, live_server):
    """This fixture is a helper that renders and locates your component in
    playwright.

    Use it in conjunction with tetra.tests.views.render_component_view and make sure
    your tests' urls.py includes a line like:

    ```python
    urlpatterns = [ path("", include("tetra.tests.urls")) ]
    ```
    """
    test_ui = Library("test_ui", "main")

    def _component_locator(component_cls, **kwargs):
        if not hasattr(component_cls, "_library") or component_cls._library is None:
            component_cls = test_ui.register(component_cls)
            test_ui.build()
        else:
            # ensure the library is built if it hasn't been yet
            # (though normally libraries registered at module level are built by tetrabuild)
            pass

        library = component_cls._library
        component_name = component_cls.__name__
        full_component_name = component_cls.full_component_name()

        component_tag = f"{{% {library.name}.{component_name} "
        for key, value in kwargs.items():
            component_tag += f'{key}="{value}" '
        component_tag += "/ %}"

        page.goto(
            live_server.url
            + reverse("render_component_view")
            + f"?component_tag={component_tag}"
        )
        return page.locator(f'[tetra-component="{full_component_name}"]')

    return _component_locator


@pytest.fixture
async def tetra_ws_communicator(db) -> AsyncGenerator[WebsocketCommunicator, Any]:
    """
    Returns a WebsocketCommunicator for TetraConsumer with a valid session and
    AnonymousUser, and closes it after usage.
    """
    from django.contrib.auth.models import AnonymousUser
    from django.contrib.sessions.backends.db import SessionStore
    from tetra.consumers import TetraConsumer

    session = SessionStore()
    session.create()

    scope = {
        "type": "websocket",
        "session": session,
        "user": AnonymousUser(),
        "path": "/ws/tetra/",
    }

    communicator = WebsocketCommunicator(
        TetraConsumer.as_asgi(),
        "/ws/tetra/",
    )
    communicator.scope.update(scope)
    connected, _ = await communicator.connect()
    assert connected
    yield communicator
    await communicator.disconnect()


@pytest.fixture
def component_render(client):
    """Fixture that renders a component class for testing"""
    test_ui = Library("test_ui", "main")

    def _component_render(component_cls: type[BasicComponent], **kwargs):
        if not hasattr(component_cls, "_library") or component_cls._library is None:
            component_cls = test_ui.register(component_cls)
            test_ui.build()
        else:
            # ensure the library is built if it hasn't been yet
            # (though normally libraries registered at module level are built by tetrabuild)
            pass

        library = component_cls._library
        component_name = component_cls.__name__
        full_component_name = component_cls.full_component_name()

        component_tag = f"{{% {library.name}.{component_name} "
        for key, value in kwargs.items():
            component_tag += f'{key}="{value}" '
        component_tag += "/ %}"

        return client.get(
            reverse("render_component_view") + f"?component_tag={component_tag}"
        )

    return _component_render


@pytest.fixture
def tetra_request() -> HttpRequest:
    """
    Returns an Http GET Request with a session, and a `tetra` attribute.
    """
    factory = RequestFactory()
    request = factory.get("/")

    add_session_to_request(request)

    request.tetra = TetraDetails(request)

    return request



import itertools
import os
import shutil
import glob
from django.conf import settings
from tetra import Library


def build(libs_to_build, force=False):
    # Clear old static files before building
    if force and hasattr(settings, "STATIC_ROOT") and settings.STATIC_ROOT:
        tetra_pattern = os.path.join(settings.STATIC_ROOT, "*/tetra/")
        for tetra_dir in glob.glob(tetra_pattern):
            if os.path.exists(tetra_dir):
                print(f"Clearing old static files: {tetra_dir}")
                shutil.rmtree(tetra_dir)

    print("Tetra: Building Javascript and CSS")
    print(" - Libraries: %s" % ",".join(o.display_name for o in libs_to_build))
    for lib in libs_to_build:
        lib.build(force=force)


def runserver_build():
    libs_to_build = list(
        itertools.chain.from_iterable(d.values() for d in Library.registry.values())
    )
    build(libs_to_build)



import hashlib
import logging
import importlib
import os
from datetime import datetime, date, time
from decimal import Decimal
import warnings

from copy import copy
from typing import Optional, Self, Any
from types import FunctionType, NoneType
from enum import Enum
import inspect
import re
import itertools
from weakref import WeakKeyDictionary
from functools import wraps
from threading import local

from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.core.files import File
from django.core.files.uploadedfile import UploadedFile
from django.db import models
from django.db.models import Model
from django.db.models.base import ModelBase
from django.forms import (
    Form,
    modelform_factory,
    BaseForm,
    FileField,
    ModelForm,
    ModelChoiceField,
)
from django.template.base import Template
from django.template.loader import render_to_string
from django.template import (
    RequestContext,
    TemplateSyntaxError,
    TemplateDoesNotExist,
    Engine,
)
from django.template.loader_tags import BLOCK_CONTEXT_KEY, BlockContext, BlockNode
from django.utils.safestring import mark_safe, SafeString
from django.utils.html import escapejs, escape
from django.utils.functional import SimpleLazyObject
from django.http import JsonResponse, HttpRequest, FileResponse, HttpResponse
from django.urls import reverse
from django.template.loaders.filesystem import Loader as FileSystemLoader

from ..exceptions import ComponentError
from ..middleware import TetraHttpRequest
from ..utils import (
    camel_case_to_underscore,
    to_json,
    TetraJSONEncoder,
    isclassmethod,
    param_names_exist,
    param_count,
    has_single_root,
    NamedTemporaryUploadedFile,
)
from ..types import ComponentData
from ..state import encode_component, decode_component
from ..templates import InlineOrigin, InlineTemplate

from .callbacks import CallbackList


thread_local = local()

logger = logging.getLogger(__name__)


def make_template(cls) -> Template:
    """Create a template from a component class.

    Uses either the `cls.template` attribute as inline template string,
    or the html template file source in the component's directory. If both are defined,
    'template' overrides 'template_name'.
    """
    from ..templatetags.tetra import get_nodes_by_type_deep

    def prepare_template_source(source):
        if re.match(r"^\s*{%\s*extends", source):
            # If the template starts with an extends tag, we must insert the
            # load tetra tag AFTER the extends tag.
            return re.sub(
                r"^(\s*{%\s*extends\s+['\"][^'\"]+['\"]\s*%})",
                r"\1{% load tetra %}",
                source,
            )
        return "{% load tetra %}" + source

    # if only "template" is defined, use it as inline template string.
    if hasattr(cls, "template"):
        if not cls.template:
            raise ComponentError(f"Component '{cls.__name__}' has an empty template.")
        making_lazy_after_exception = False
        filename, line = cls.get_template_source_location()
        origin = InlineOrigin(
            name=f"{filename}:{cls.__name__}.template",
            template_name=filename,
            start_line=line,
            component=cls,
        )
        try:
            template = InlineTemplate(
                prepare_template_source(cls.template),
                origin=origin,
            )
        except TemplateSyntaxError as e:
            # By default, we want to compile templates during python compile time,
            # however, the template exceptions are much better when raised at runtime
            # as it shows a nice stack trace in the browser. We therefore create a
            # "Lazy" template after a compile error that will run in the browser when
            # testing.
            from django.conf import settings

            logger.error(
                f"Failed to compile inline template for component '{cls.__name__}': {e}"
            )
            if settings.DEBUG:
                making_lazy_after_exception = True
                template = SimpleLazyObject(
                    lambda: InlineTemplate(
                        prepare_template_source(cls.template),
                        origin=origin,
                    )
                )
            else:
                raise ComponentError(
                    f"Template compilation failed for component '{cls.__name__}': {e}"
                )

        if not making_lazy_after_exception and template is not None:
            for i, block_node in enumerate(
                get_nodes_by_type_deep(template.nodelist, BlockNode)
            ):
                if not getattr(block_node, "origin", None):
                    block_node.origin = origin

    elif hasattr(cls, "template_name"):
        raise NotImplementedError(
            f"'template_name' is not implemented for Tetra components ({cls})"
        )
        # template_file_name = cls.template_name
        # template_dir = os.path.dirname(template_file_name)

    else:

        # try to find <component_name>.html within component's directory
        module = importlib.import_module(cls.__module__)

        # No template, no template_name attrs found
        # Here we definitely must have a dir-style component
        if not hasattr(module, "__path__"):
            raise ComponentError(
                f"'{cls.__module__}.{cls.__name__}' is not a valid component. "
                f"You either have to put it into a correct library directory with a "
                f"template HTML alongside, or add a 'template'/'template_name' "
                f"attribute to the component."
            )
        module_path = module.__path__[0]
        component_name = module.__name__.split(".")[-1]
        # if path is a file, get the containing directory
        # template_dir = os.path.dirname(module_path)

        # FIXME: better use cls.get_template_source_location() for this!
        # template_source = cls._read_component_file_with_extension("html")

        template_file_name = f"{component_name}.html"

        # Load the template using a custom loader
        try:
            engine = Engine(dirs=[module_path], app_dirs=False)
            loader = FileSystemLoader(engine)
            for template_path in loader.get_template_sources(template_file_name):
                try:
                    # Open and read the template source
                    with open(template_path.name, "r", encoding="utf-8") as f:
                        template_source = f.read()

                    origin = InlineOrigin(
                        name=os.path.join(module_path, template_file_name),
                        template_name=template_file_name,
                        start_line=0,
                        component=cls,
                    )
                    # Compile the template
                    template = Template(
                        prepare_template_source(template_source),
                        origin,
                        template_file_name,
                    )
                    break
                except FileNotFoundError:
                    # If the file is not found, continue with the next source
                    continue
            else:
                # If no template is found, raise an error
                raise ComponentError(
                    f"Template file '{template_file_name}' not found for component"
                    f" '{cls.__name__}'."
                )
            # template = get_template(template_file_name).template
        except TemplateDoesNotExist:
            raise ComponentError(
                f"Template file '{template_file_name}' not found for component"
                f" '{cls.__name__}'."
            )
    if not re.match(r"^\s*{%\s*extends", template.source) and not has_single_root(
        template.source
    ):
        raise ComponentError(
            f"Component template '{cls.__name__}.template' must contain exactly "
            f"one top-level tag."
        )
    return template


class RenderDataMode(Enum):
    """The mode how to render the component's data."""

    INIT = 0  # initialize data. There was no component state before
    MAINTAIN = 1  # keep component data
    UPDATE = 2  # update component data with new data from server


class BasicComponent:
    __abstract__ = True
    style: str = ""
    component_id: Optional[str] = None
    _is_resumed_from_state = False
    _is_directory_component: bool = False
    _template: Template
    _to_compile = []
    _name = None
    _library = None
    _app = None

    def __init_subclass__(cls, **kwargs):
        from tetra.state import loading_libraries

        super().__init_subclass__(**kwargs)
        cls._name = camel_case_to_underscore(cls.__name__)

        if "__abstract__" not in cls.__dict__ or cls.__dict__["__abstract__"] is False:
            if loading_libraries:
                # postpone template compiling to time when all libraries are loaded
                cls._to_compile.append(cls)
            else:
                cls._template = make_template(cls)

    def __init__(
        self,
        _request: TetraHttpRequest | HttpRequest,
        _attrs: dict[str, Any] | None = None,
        _context: dict[str, Any] | RequestContext | None = None,
        _slots=None,
        *args,
        **kwargs,
    ) -> None:
        """Initializes the component with the provided attributes and context.

        It calls load() with the optional args/kwargs."""

        super().__init__()  # call init without kwargs.
        self.request = _request
        self.attrs = _attrs or {}
        self._context = _context or {}
        self._slots = _slots
        # FIXME: it could lead to mismatching component ids if it is recreated after
        #  page reloading - test this for channels/long-lasting websocket connections
        self.component_id = self.get_component_id()
        self._call_load(*args, **kwargs)

    @classmethod
    def _compile_all_templates(mcls):
        while mcls._to_compile:
            cls = mcls._to_compile.pop(0)
            if not hasattr(cls, "_template"):
                cls._template = make_template(cls)

    def get_component_id(self) -> str:
        """Creates a unique, reproducible, session-persistent component id.

        This is calculated from the component name, optionally the session key,
        and the component key, if available.
        """
        try:
            name = self.full_component_name()
        except ComponentError:
            name = f"{self.__class__.__module__}.{self.__class__.__name__}"

        session_key = (
            self.request.session.session_key if hasattr(self.request, "session") else ""
        )
        component_key = self.attrs.get("key", "")
        s = f"{name}{session_key}{component_key}"
        return hashlib.blake2b(s.encode(), digest_size=8).hexdigest()

    @classmethod
    def full_component_name(cls) -> str:
        if not cls._library:
            raise ComponentError(
                f"Component class '{cls.__name__}' is not registered in a library."
            )
        return f"{cls._library.app.label}__{cls._library.name}__{cls._name}"

    def get_extra_tags(self) -> dict[str, str | None]:
        """Returns extra tags to be included in the component's HTML.

        This method can be overridden in subclasses to add custom
        attributes or classes to the component's HTML. Don't forget to call
        super().get_extra_tags() in your own implementation.
        """
        return {"tetra-component-id": self.component_id}

    @classmethod
    def get_source_location(cls) -> tuple[str, int, int]:
        """Returns the filename, line number, and line count of the component's
        source code."""
        filename = inspect.getsourcefile(cls)
        if not filename:
            raise ComponentError(
                f"Could not find source file for component '{cls.__name__}'."
            )
        lines, start = inspect.getsourcelines(cls)
        return filename, start, len(lines)

    @classmethod
    def get_template_source_location(cls) -> tuple[str, int | None]:
        """Returns the filename and startline number of the component's template,
        if available."""
        filename, comp_start, com_end = cls.get_source_location()
        if not hasattr(cls, "template") or not cls.template:
            # this is a directory style component
            filename = cls._get_component_file_path_with_extension("html")
            return filename, 0
        with open(filename, "r") as f:
            source = f.read()
        try:
            start = source.index(cls.template)
        except ValueError:
            # If the template is not found exactly, it might be due to
            # indentation or other formatting changes if the component
            # was defined dynamically.
            return filename, comp_start
        line = source[:start].count("\n") + 1
        return filename, line

    @classmethod
    def _get_component_file_path_with_extension(cls, extension):
        if cls._is_directory_component is not True:
            # assume this is an inline component, no css/js files available
            return ""
        module = importlib.import_module(cls.__module__)
        component_name = module.__name__.split(".")[-1]
        module_path = module.__path__[0]
        file_name = f"{component_name}.{extension}"
        return os.path.join(module_path, file_name)

    @classmethod
    def _read_component_file_with_extension(cls, extension):
        file_path = cls._get_component_file_path_with_extension(extension)
        if not os.path.exists(file_path):
            return ""

        try:
            with open(file_path, "r") as f:
                return f.read()
        except Exception as e:
            # handle al other errors as well: IOError, UnicodeDecodeError, OSError
            # FileNotFoundError is already handled by exists() above
            logger.critical(f"Error reading component file '{file_path}': {e}")
            return ""

    @classmethod
    def has_script(cls) -> bool:
        return False

    @classmethod
    def render_script(cls, component_var=None) -> str:
        """In BasicComponent, always returns an empty string."""
        return ""

    @classmethod
    def has_styles(cls) -> bool:
        """Returns True if the component has a css style defined in the class or a file
        in the component directory."""
        if cls.style:
            return True
        else:
            return os.path.exists(cls._get_component_file_path_with_extension("css"))

    @classmethod
    def _is_styles_inline(cls) -> bool:
        return bool(cls.style)

    @classmethod
    def extract_styles(cls) -> str:
        """Returns the filename and whether the style was found in the component's source code.

        Returns:
            filename: str
            found: bool
        """
        # check if we have a style defined in the class, otherwise check if there is
        # a file in the component directory
        if cls._is_styles_inline():
            source_filename, comp_start_line, source_len = cls.get_source_location()
            with open(source_filename, "r") as f:
                py_source = f.read()
            comp_start_offset = len("\n".join(py_source.split("\n")[:comp_start_line]))
            start = py_source.index(cls.style, comp_start_offset)
            before = py_source[:start]
            before = re.sub(r"\S", " ", before)
            return f"{before}{cls.style}"
        else:
            return cls._read_component_file_with_extension("css")

    @classmethod
    def render_styles(cls) -> str:
        """Returns the CSS styles defined in the class inline, or from a component's
        CSS file."""

        # for CSS, this is nothing else than returning the styles as string.
        # there is no additional logic here.
        return cls.extract_styles()

    @classmethod
    def as_tag(cls, request, *args, **kwargs) -> SafeString:
        if not hasattr(request, "tetra_components_used"):
            request.tetra_components_used = set()
        request.tetra_components_used.add(cls)
        component = cls(request, *args, **kwargs)
        component.recalculate_attrs(component_method_finished=False)
        return component.render()

    def _call_load(self, *args, **kwargs) -> None:
        self._pre_load(*args, **kwargs)
        self.load(*args, **kwargs)

    def _pre_load(self, *args, **kwargs):
        """Placeholder for code that should be run right before load() is called.
        This can be used for inheriting base classes to auto-load some data,
        but users don't have to call super().load() from their load() implementation
        each time.
        """
        pass

    def load(self, *args, **kwargs) -> None:
        """Override this method to load the component's data, e.g. from the database.

        Caution: Any attributes that are set in this method are NOT saved with the
        state, for efficiency.
        """
        pass

    def _add_self_attrs_to_context(self, context) -> None:
        for key in dir(self):
            if not (
                key.startswith("_") or isclassmethod(getattr(self.__class__, key, None))
            ):
                # if not (key.startswith("_") or isclassmethod(getattr(self, key))):
                context[key] = getattr(self, key)

    def get_context_data(self, **kwargs) -> RequestContext:
        """Update the render context of the component with given kwargs.

        Make sure to call the `super().get_context_data()` method when overriding.
        """
        if isinstance(self._context, RequestContext):
            context = self._context
        else:
            context = RequestContext(self.request, self._context)
        self._add_self_attrs_to_context(context)
        context.update(kwargs)
        return context

    def render(self) -> SafeString:
        self.recalculate_attrs(component_method_finished=True)
        context = self.get_context_data()

        with context.render_context.push_state(self._template, isolated_context=True):
            if self._slots:
                if BLOCK_CONTEXT_KEY not in context.render_context:
                    context.render_context[BLOCK_CONTEXT_KEY] = BlockContext()
                block_context = context.render_context[BLOCK_CONTEXT_KEY]
                block_context.add_blocks(self._slots)

            if context.template is None:
                with context.bind_template(self._template):
                    html = self._template._render(context)
            else:
                html = self._template._render(context)

        return mark_safe(html)

    def recalculate_attrs(self, component_method_finished: bool):
        """Hook for code that should be run before and after user interactions with
        attributes.

        You can add code here that e.g. updates some attributes "in the last moment"
        before rendering, like "dirty" flags that are set when any of the attributes
        have changed, or property attributes that are calculated automatically,
        after component methods have changed other attributes.

        The method is called before and after the component methods are executed.

        Attributes:
            component_method_finished (bool): Whether the recalculation was triggered
                before or after (right before rendering) custom component methods are
                executed. You can react on it differently.
        """


empty = object()


class PublicMeta(type):
    def __getattr__(self, name):
        do_name = f"do_{name}"
        if do_name in self.__dict__:
            inst = self()
            return getattr(inst, do_name)
        else:
            raise AttributeError(f"Public decorator has no method {name}.")


class Public(metaclass=PublicMeta):
    def __init__(self, obj: Any = None, update: bool = True) -> None:
        """
        Decorate a method or attribute with a public decorator.

        This decorator can be used to create methods or attributes that behave like
        public methods or properties. It can be used to add event listeners, watch
        attributes, debounce and throttle functions, and other functionality.

        Example usage:
        class MyComponent(Component):
            my_attribute = public("initial value")

            @public
            def my_method(self):
                ...

            @public
            @watch("my_attribute")
            def my_attribute_changed(self, new_value, old_value, attr):
                print(f"My attribute changed from {old_value} to {new_value}")

        my_component = MyComponent()
        my_component.my_method()  # Output: My method called

        Attributes:
            obj (Any): The object to decorate (method or attribute).
            update (bool): whether update() should be called at the end of the method.
        """
        self._update = update
        self._watch: list[str] = []
        self._debounce = None
        self._debounce_immediate = None
        self._throttle = None
        self._throttle_trailing = None
        self._throttle_leading = None
        self._event_subscriptions: list[str] = []
        self.obj = None
        self.__call__(obj)

    def __call__(self, obj: Any) -> Self:
        if isinstance(obj, Public):
            # Public decorator applied multiple times - combine them
            self._update = obj._update if obj._update else self._update
            self._watch = obj._watch if obj._watch else self._watch
            self._debounce = obj._debounce if obj._debounce else self._debounce
            self._debounce_immediate = (
                obj._debounce_immediate
                if obj._debounce_immediate
                else self._debounce_immediate
            )
            self._throttle = obj._throttle if obj._throttle else self._throttle
            self._throttle_trailing = (
                obj._throttle_trailing
                if obj._throttle_trailing
                else self._throttle_trailing
            )
            self._throttle_leading = (
                obj._throttle_leading
                if obj._throttle_leading
                else self._throttle_leading
            )
            self._event_subscriptions = (
                obj._event_subscriptions
                if obj._event_subscriptions
                else self._event_subscriptions
            )
            self.obj = obj.obj if obj.obj else self.obj

        elif isinstance(obj, FunctionType):
            # `public` decorator applied to a method

            @wraps(obj)
            def fn(instance: "Component", *args, **kwargs):
                ret: JsonResponse | FileResponse = obj(instance, *args, **kwargs)
                if self._update:
                    instance.update()
                return ret

            self.obj = fn
        else:
            # `public` is wrapping a variable (str, int, etc)
            self.obj = obj
        return self

    def __getattr__(self, name) -> Any:
        """If an attribute name is requested that does not exist in the class,
        search for `do_<name>` in the class.
        """
        do_name = f"do_{name}"
        if do_name in self.__class__.__dict__:
            return getattr(self, do_name)
        else:
            raise AttributeError(f"Public decorator has no method {name}.")

    def do_watch(self, *args) -> Self:
        if not args:
            raise ValueError(".watch decorator requires at least one argument.")
        for arg in args:
            if isinstance(arg, str):
                self._watch.append(arg)
            else:
                self._watch.extend(arg)
        return self

    def do_debounce(self, timeout, immediate=False) -> Self:
        self._debounce = timeout
        self._debounce_immediate = immediate
        return self

    def do_throttle(self, timeout, trailing=False, leading=True) -> Self:
        self._throttle = timeout
        self._throttle_trailing = trailing
        self._throttle_leading = leading
        return self

    def do_subscribe(self, event) -> Self:
        """Deprecated. Use do_listen instead."""
        warnings.warn(
            "@public.subscribe is deprecated. Use @public.listen instead.",
            DeprecationWarning,
        )
        return self.do_listen(event)

    def do_listen(self, event) -> Self:
        """Keeps track of the event for a dynamic event subscription of the component."""
        self._event_subscriptions.append(event)
        return self


public = Public


tracing_component_load = WeakKeyDictionary()


class ComponentMetaClass(type):
    def __new__(mcls, name, bases, attrs):
        public_methods: list[dict[str, Any]] = list(
            itertools.chain.from_iterable(
                base._public_methods
                for base in bases
                if hasattr(base, "_public_methods")
            )
        )
        public_properties: list[str] = list(
            itertools.chain.from_iterable(
                base._public_properties
                for base in bases
                if hasattr(base, "_public_properties")
            )
        )
        for attr_name, attr_value in attrs.items():

            if isinstance(attr_value, Public):
                # if there is public decorated attribute/method, replace it with the
                # attribute/method itself, as we don't need the decorator anymore.

                attrs[attr_name] = attr_value.obj

                if isinstance(attrs[attr_name], FunctionType):
                    # the decorated object is a method
                    fn = attrs[attr_name]
                    pub_met = {"name": attr_name}
                    if attr_value._watch:
                        if not param_names_exist(fn, "value", "old_value", "attr"):
                            raise ValueError(
                                f"The .watch method `{attr_name}` must have 'value', "
                                f"'old_value' and 'attr' as arguments."
                            )
                        pub_met["watch"] = attr_value._watch
                    if attr_value._debounce:
                        pub_met["debounce"] = attr_value._debounce
                        pub_met["debounce_immediate"] = attr_value._debounce_immediate
                    if attr_value._throttle:
                        pub_met["throttle"] = attr_value._throttle
                        pub_met["throttle_trailing"] = attr_value._throttle_trailing
                        pub_met["throttle_leading"] = attr_value._throttle_leading
                    if attr_value._event_subscriptions:
                        pcount = param_count(fn)
                        if pcount != 1:
                            raise ValueError(
                                f"Event listener method '{attr_name}' has wrong "
                                f"number of arguments. Expected 2 (self, "
                                f"event_detail), but got {pcount}."
                            )
                        pub_met["event_subscriptions"] = attr_value._event_subscriptions
                    public_methods.append(pub_met)
                else:
                    public_properties.append(attr_name)
        newcls = super().__new__(mcls, name, bases, attrs)
        newcls._public_methods = public_methods
        newcls._public_properties = public_properties
        # Reset registration info inherited from base classes
        if "_library" in attrs:
            newcls._library = attrs["_library"]
        else:
            newcls._library = None

        if "_name" in attrs:
            newcls._name = attrs["_name"]
        else:
            newcls._name = None
        return newcls


class Component(BasicComponent, metaclass=ComponentMetaClass):
    __abstract__ = True
    script: Optional[str] = None
    _callback_queue = None
    _excluded_props_from_saved_state = [
        "request",
        "_callback_queue",
        "_loaded_children_state",
        "_excluded_load_props_from_saved_state",
        "_is_resumed_from_state",
        "_resumed_from_state_data",
        "_event_subscriptions",
        "__abstract__",
        "_temp_files",
        "_is_directory_component",
    ]
    _excluded_load_props_from_saved_state: list[str] = []
    _loaded_children_state = None
    _load_args: list[str] = []
    _load_kwargs: dict[str, Any] = {}
    _resumed_from_state_data: ComponentData
    # _temp_files is an internal dict to track which data attributes are files.
    _temp_files: dict[str, UploadedFile] = {}
    key = public(None)

    def __init__(self, _request, key=None, *args, **kwargs) -> None:
        super().__init__(_request, *args, **kwargs)
        self.key = key if key is not None else self.full_component_name()

    @classmethod
    def from_state(
        cls,
        component_state: ComponentData,
        request: HttpRequest,
        key: str | None = None,
        _attrs=None,
        _context=None,
        _slots=None,
        *args,
        **kwargs,
    ) -> Any:
        """Creates and initializes a component instance from its serialized state.

        This method reconstructs a component from its encrypted state data, typically
        received from the client after a previous server interaction. It handles:

        - Decrypting and validating the component state
        - Restoring component attributes and properties
        - Resolving Django Model instances from their primary keys
        - Managing uploaded files and their temporary storage
        - Re-executing the component's load() method with original arguments
        - Updating public properties from client-side changes

        Args:
            component_state: Dictionary containing 'encrypted' state string and 'data'
                dict with current client-side property values
            request: The current HTTP request object
            key: Optional component key for identification
            _attrs: Optional attributes to set on the component
            _context: Optional template context to use
            _slots: Optional template slots to use
            *args: Positional arguments to pass to load()
            **kwargs: Keyword arguments to pass to load()

        Returns:
            Component: A fully initialized component instance with restored state

        Raises:
            TypeError: If component_state structure is invalid
            AttributeError: If required state attributes are missing

        Note:
            - Model fields are automatically fetched from the database using their PKs
            - File fields maintain references to temporary uploaded files
            - The component's load() method is called with the original arguments
            - Client-side property changes override the encrypted state values
        """
        if not isinstance(component_state, dict):
            raise TypeError("Encrypted state has wrong type")
        if "encrypted" not in component_state:
            raise AttributeError("No 'encrypted' attribute in component_state")
        if not isinstance(component_state["encrypted"], str):
            raise TypeError(
                f"Encrypted state has wrong type: {type(component_state['encrypted'])}"
            )
        if "data" not in component_state:
            raise AttributeError("No 'data' attribute in encrypted state")
        if not isinstance(component_state["data"], dict):
            raise TypeError("Invalid component state 'data' type.")

        component = decode_component(component_state["encrypted"], request)
        if not isinstance(component, cls):
            raise TypeError(
                f"Component '{component.__class__.__name__}' of invalid "
                f"type, should be: {cls.__name__}."
            )

        component.request = request
        # Set default values for attributes if they don't exist
        for attr, default in [
            ("_context", {}),
            ("attrs", {}),
            ("_temp_files", {}),
            ("_slots", None),
        ]:
            if not hasattr(component, attr):
                setattr(component, attr, default)

        # Override with provided values if given
        if key:
            component.key = key
        if _attrs:
            component.attrs = _attrs
        if _context:
            component._context = _context
        if _slots:
            component._slots = _slots
        if args:
            component._load_args = args
        if kwargs:
            component._load_kwargs = kwargs
        component._recall_load()

        # get data from client and populate component attributes with it
        for key, state_value in component_state["data"].items():
            # try to get attribute type (from the class annotations created when the
            # component was declared)
            AttributeType = component.__annotations__.get(key, NoneType)

            # if client data type is a model, try to see the value in the
            # recovered data as Model.pk and get the model again.
            if issubclass(AttributeType, Model):
                if state_value:
                    # FIXME: this hits the database, before the actual (and probably
                    #  different) value is set via e.g. the load() method.
                    #  Find a way to only load the model when needed.
                    state_value = AttributeType.objects.get(pk=state_value)
                else:
                    state_value = None  # or attr_type.objects.none()

            # if the data type is a file, try to find the file by its path
            elif issubclass(AttributeType, UploadedFile):
                if state_value:
                    # if the file was reconstructed from the client data (which
                    # doesn't contain the temp_path anymore), recover the temp_path
                    # from the original state (decoded from the encrypted token)
                    if getattr(state_value, "_reconstructed", False):
                        original_value = getattr(component, key, None)
                        if isinstance(original_value, NamedTemporaryUploadedFile):
                            # reuse the original file handle.
                            state_value.file = original_value.file
                        else:
                            # if the file is not in the original state, it means it
                            # was not uploaded by this user in a previous request.
                            state_value = None

                    if not state_value:
                        continue

                    # the value (temp file path) from the saved component client state
                    # could be stale, as the file could be already saved as permanent
                    # file in MEDIA_ROOT/* somewhere using submit() in another session.
                    # Then the state's file_path is wrong. in this case,
                    # delete this state!
                    if not state_value.file or not os.path.exists(
                        state_value.file.name
                    ):
                        component_state["data"][key] = None
                        if key in component._temp_files:
                            del component._temp_files[key]

                else:
                    # if there is no valid file in the request, it means the file was
                    # not correctly uploaded or removed from the input field,
                    # so just skip it
                    continue

                # save file in a separate dict so we can access it easier
                component._temp_files[key] = state_value

            setattr(component, key, state_value)

        component._is_resumed_from_state = True
        component._resumed_from_state_data = component_state["data"]
        component.recalculate_attrs(component_method_finished=False)
        return component

    @classmethod
    def _component_url(cls, method_name) -> str:
        return reverse(
            "tetra_public_component_method",
            args=[cls._library.app.label, cls._library.name, cls._name, method_name],
        )

    @classmethod
    def has_script(cls) -> bool:
        """Returns True if the component has a javascript script part, else False."""

        # First check if the script is defined in the class, otherwise check if there
        # is a file in the component directory
        if cls.script:
            return True
        else:
            return os.path.exists(cls._get_component_file_path_with_extension("js"))

    @classmethod
    def _is_script_inline(cls) -> bool:
        """Returns True if the component has a JavaScript script,
        and this script is declared inline.

        If the component has a script file AND an inline script, the inline script
        takes precedence, and True is returned."""
        return bool(cls.script)

    @classmethod
    def extract_script(cls) -> str:
        """This method extracts the component's JavaScript, from wherever it finds
        it, and returns it.

        Returns:
            A tuple with the filename of the javascript script and a boolean value
                which is True if the JavaScript was found inline in the component code,
                False if there was an external .js file.
        """
        if cls._is_script_inline():
            source_filename, comp_start_line, source_len = cls.get_source_location()
            with open(source_filename, "r") as f:
                py_source = f.read()
            comp_start_offset = len("\n".join(py_source.split("\n")[:comp_start_line]))
            start = py_source.index(cls.script, comp_start_offset)
            before = py_source[:start]
            before = re.sub(r"\S", " ", before)
            return f"{before}{cls.script}"
        else:
            # Find script in the component's directory
            return cls._read_component_file_with_extension("js")

    @classmethod
    def render_script(cls, component_var=None) -> str:
        """This method dynamically generates the complete JavaScript module for the
        component.

        It consists of the component's methods, attributes, and server-side methods,
        including custom JavaScript code from the script property or a .js file in
        the component's directory.

        This script can be imported dynamically via Alpine.init() and used to update
        the component's state.
        """
        component_server_methods = []
        for method in cls._public_methods:
            method_data = copy(method)
            method_data["endpoint"] = (cls._component_url(method["name"]),)
            component_server_methods.append(method_data)

        if not component_var:
            component_var = cls.extract_script() if cls.has_script() else "{}"
        return render_to_string(
            "script.js",
            {
                "component_name": cls.full_component_name(),
                "component_script": component_var,
                "component_server_methods": to_json(component_server_methods),
                "component_server_properties": to_json(cls._public_properties),
            },
        )

    def _call_load(self, *args, **kwargs) -> None:
        """Load the component's state and attributes.
        It keeps a record of the given parameters for later recalls and
        then calls the (user defined) load method of the component.
        """
        self._load_args = args
        self._load_kwargs = kwargs
        # prevent properties set in load() to be saved with the state
        tracing_component_load[self] = set()
        try:
            self._pre_load(*args, **kwargs)
            self.load(*args, **kwargs)
            props = tracing_component_load[self]
        finally:
            del tracing_component_load[self]
        self._excluded_load_props_from_saved_state = list(props)

    def _recall_load(self) -> None:
        """Re-execute the load method of the component with the same arguments that
        were used when the component was initially loaded."""
        self._call_load(*self._load_args, **self._load_kwargs)

    def __setattr__(self, item, value) -> None:
        """Special method that allows attributes to be set on the component. It also
        tracks which attributes are being set so that they can be included in the
        component's state.
        """
        if self in tracing_component_load:
            # TODO: maybe File is too broad? FieldFile, UploadedFile?
            if not isinstance(value, File):
                tracing_component_load[self].add(item)
        return super().__setattr__(item, value)

    @property
    def client(self) -> CallbackList:
        if self._callback_queue is None:
            self._callback_queue = CallbackList()
        return self._callback_queue

    def set_load_args(self, *args, **kwargs) -> None:
        load_args = {
            "args": args,
            "kwargs": kwargs,
        }
        try:
            to_json(load_args)
        except TypeError:
            raise ComponentError(
                f"Tetra Component {self.__class__.__name__} tried to self.set_load_args() with a none json serializable value."
            )
        self._load_args = load_args

    def _data(self) -> dict[str, Any]:
        return {key: getattr(self, key) for key in self._public_properties}

    def get_extra_tags(self) -> dict[str, str | None]:
        extra_tags = super().get_extra_tags()
        extra_tags.update(
            {
                "tetra-component": f"{self.full_component_name()}",
                "x-bind": "__rootBind",
            }
        )
        return extra_tags

    def _encoded_state(self) -> str:
        return encode_component(self)

    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__.copy()
        for key in (
            self._excluded_props_from_saved_state
            + self._excluded_load_props_from_saved_state
        ):
            if key in state:
                del state[key]
        return state

    def _render_data(self) -> dict[str, Any]:
        """Returns a dictionary that includes the component's attributes and a
        special attribute __state that contains the component's encoded state as a JSON
        string.
        This dictionary is used to render the component's HTML.
        """
        data = self._data()
        data["__state"] = self._encoded_state()
        return data

    def _add_self_attrs_to_context(self, context) -> None:
        super()._add_self_attrs_to_context(context)
        if hasattr(self, "_loaded_children_state") and self._loaded_children_state:
            children_state = {c["data"]["key"]: c for c in self._loaded_children_state}
            context["_loaded_children_state"] = children_state
        else:
            context["_loaded_children_state"] = None

    def render(self, mode=RenderDataMode.INIT) -> SafeString:
        """Renders the component's HTML.

        It makes the following decisions based on the given mode:
        - If mode is RenderDataMode.INIT (the default), just renders the component's
            data into `x-data`. The client will use it as Alpine data.
        - If mode is RenderDataMode.MAINTAIN, it instructs the client to just copy the
            original `x-data` to the updated component element, so no data changes.
        - If mode is RenderDataMode.UPDATE, the client will be instructed to replace
            the existing `x-data` with new data from the server state
        """
        if hasattr(thread_local, "_tetra_render_mode"):
            mode = thread_local._tetra_render_mode
            set_thread_local = False
        else:
            thread_local._tetra_render_mode = mode
            set_thread_local = True
        html = super().render()
        if set_thread_local:
            del thread_local._tetra_render_mode
        tag_name = re.match(r"^\s*(<!--.*-->)?\s*<\w+", html)
        if not tag_name:
            # If there is an extends tag, the rendered HTML might not start with
            # a tag if it's just blocks being filled. However, Tetra components
            # are expected to have a root element.
            # When extending, the root element should be in the base template.
            # But the regex check is on the *rendered* html.
            if not re.match(r"^\s*{%\s*extends", self._template.source):
                raise ComponentError(
                    f"Error in {self.__class__.__name__}: The component's template is "
                    "not enclosed in HTML tags."
                )
            # If it is an extends template, we should probably find the first tag
            # in the rendered output, which might have come from the base template.
            tag_name = re.search(r"<\w+", html)
            if not tag_name:
                raise ComponentError(
                    f"Error in {self.__class__.__name__}: The component's rendered "
                    "template is not enclosed in HTML tags."
                )

        tag_name_end = tag_name.end(0)

        extra_tags = self.get_extra_tags()

        if self.key:
            extra_tags.update({"key": self.key})
        if mode == RenderDataMode.UPDATE and self._is_resumed_from_state:
            data_json = escape(to_json(self._render_data()))
            old_data_json = escape(to_json(self._resumed_from_state_data))
            extra_tags.update({"x-data": ""})
            extra_tags.update({"x-data-update": data_json})
            extra_tags.update({"x-data-update-old": old_data_json})
        elif mode == RenderDataMode.MAINTAIN:
            extra_tags.update({"x-data": ""})
            extra_tags.update({"x-data-maintain": ""})
        else:
            data_json = escapejs(to_json(self._render_data()))
            extra_tags.update(
                {"x-data": f"{self.full_component_name()}('{data_json}')"}
            )

        for method_data in self._public_methods:
            if "event_subscriptions" in method_data:
                for event in method_data["event_subscriptions"]:
                    extra_tags.update(
                        {f"@{event}": f'{method_data["name"]}($event.detail)'}
                    )
        tags_strings = [f"{key}=\"{value or ''}\"" for key, value in extra_tags.items()]
        html = f'{html[:tag_name_end]} {" ".join(tags_strings)} {html[tag_name_end:]}'
        return mark_safe(html)

    def update_html(self, include_state=False) -> None:
        """Updates the component's HTML.

        If `include_state` is True, it includes the component's state in the HTML.
        Otherwise, it only includes the component's attributes and instructs the
        client to not change the client side data.
        """
        if include_state:
            self.client._updateHtml(self.render(mode=RenderDataMode.UPDATE))
        else:
            self.client._updateHtml(self.render(mode=RenderDataMode.MAINTAIN))

    def update_data(self) -> None:
        """Updates the component's state with the latest data from the server."""
        self.client._updateData(self._render_data())

    def update(self) -> None:
        """Updates the component's HTML and state."""
        self.update_html(include_state=True)

    def replace_component(self) -> None:
        """Replaces the current component with a new one. It first creates a new
        component with the same name and attributes as the current component and
        inserts it into the DOM, and then removes the current component from the DOM.
        """
        self.client._replaceComponent(self.render())

    def push_url(self, path: str) -> None:
        """Pushes a new URL path to the browser's history."""
        self.client._pushUrl(path)
        self.request.tetra.set_url_path(path)

    def update_search_param(self, param: str, value: str | int = None):
        """Updates a URL query parameter. Leaves alone tho other params, if any.

        Args:
            param: The name of the URL search parameter to be updated.
            value: The new value for the parameter. Defaults to None, which means that
                the parameter is deleted from the URL.
        """
        self.client._updateSearchParam(param, value)
        self.request.tetra.set_url_query_param(param, value)

    def replace_url(self, url: str) -> None:
        """Replaces the current URL with a new one."""
        self.client._pushUrl(url, replace=True)
        self.request.tetra.set_url(url)

    def _call_public_method(
        self, request, method_name, children_state, *args
    ) -> JsonResponse | FileResponse:
        self._loaded_children_state = children_state
        self._callback_queue = CallbackList()
        result = getattr(self, method_name)(*args)
        callbacks = self._callback_queue.serialize()
        self._callback_queue = None
        libs = list(
            set(component._library for component in request.tetra_components_used)
        )

        # if the response is a FileResponse, we directly return the result.
        if isinstance(result, FileResponse):
            result.headers["T-Response"] = "true"
            result.as_attachment = True
            return result
        else:
            # TODO: error handling
            # TODO: use TetraMessage for this, for more consistence!
            return JsonResponse(
                {
                    "styles": [lib.styles_url for lib in libs],
                    "js": [lib.js_url for lib in libs],
                    "success": True,
                    "result": result,
                    "callbacks": callbacks,
                },
                encoder=TetraJSONEncoder,
                headers={"T-Response": "true"},
            )

    @public
    def _refresh(self) -> None:
        """Re-render and return
        This is just a noop as the @public decorator implements this functionality
        """


def get_python_type_from_form_field(field, initial) -> type:
    """
    Derives the Python type from a Django form field.

    Args:
        field: A Django form field instance

    Returns:
        The corresponding Python type. If it can't be determined, return NoneType
    """
    # The order of this map is important, as some fields derive from others. If we
    # checked a field against IntegerField before checking agains FloatField,
    # it would be True, even if it was a FloatField. We could have checked with "is"
    # or "==", but then we would have lost the ability to handle future or
    # existing field types derived from these base types.
    field_type_map = {
        forms.FloatField: float,
        forms.DecimalField: Decimal,
        forms.IntegerField: int,
        forms.BooleanField: bool,
        forms.DateField: date,
        forms.DateTimeField: datetime,
        forms.TimeField: time,
        forms.EmailField: str,
        forms.URLField: str,
        forms.CharField: str,
        forms.ChoiceField: str,
        forms.FileField: UploadedFile,
        forms.ImageField: UploadedFile,
    }

    # Check for ModelChoiceField first (special case)
    if isinstance(field, ModelChoiceField):
        return field.queryset.model

    # Check against field type map
    for field_class, python_type in field_type_map.items():
        if isinstance(field, field_class):
            return python_type

    # Fallback: try to infer from initial value
    if initial is not None:
        initial = field.to_python(initial)
        return type(initial)
    else:
        # Infer type from field class
        initial = field.to_python(initial or "")
        return type(initial) if initial is not None else NoneType


class FormComponentMetaClass(ComponentMetaClass):
    def __new__(cls, name, bases, dct):
        # iter through form fields if there is a static form_class, and set a public
        # attribute to the component for each field

        form_class = dct.get("form_class", None)
        dct.setdefault("__annotations__", {})

        if form_class:
            for field_name, field in form_class.base_fields.items():
                initial = field.initial
                # try to read the type of the component attribute from the form field
                python_type = get_python_type_from_form_field(field, initial)

                # logger.debug(f"  {field_name} -> {python_type}")
                dct[field_name] = public(initial)
                dct["__annotations__"][field_name] = python_type
        return super().__new__(cls, name, bases, dct)


class FormComponent(Component, metaclass=FormComponentMetaClass):
    """
    Component that can render a form, validate and submit it.

    The form itself is not saved with the client state, as pickling forms is somehow
    complicated. Instead, it is recreated with every request and filled with the
    current component attribute values.

    Attributes:
        form_class (type(BaseForm)):
            The form class to use for this component. All fields of the form are
            dynamically created as public attributes in the component,
            and initialized with the form's fields' initial values.
    """

    __abstract__ = True
    form_class: Optional[type[forms.ModelForm]] = None
    form_submitted: bool = False
    form_errors: dict = {}  # TODO: make protected + include in render context
    _form: Form = None
    _validate_called: bool = False

    _excluded_props_from_saved_state = Component._excluded_props_from_saved_state + [
        "_form",
        "_validate_called",
        "form_errors",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.form_class:
            raise AttributeError(
                f"Error in {self.__class__.__name__}: The 'form_class' attribute is not set."
            )

    def recalculate_attrs(self, component_method_finished: bool):
        if component_method_finished:
            # Don't show form errors if form is not submitted yet
            if not self.form_submitted:
                self._form.errors.clear()
            # If form was submitted, ensure it's validated so errors are available
            else:
                # Trigger validation if not already done
                _ = self._form.errors
        else:
            self._form = self.get_form()

    def get_context_data(self, **kwargs) -> RequestContext:
        if "form" not in kwargs:
            kwargs["form"] = self._form
        return super().get_context_data(**kwargs)

    def get_form_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        # Exclude file fields from data dict to prevent file information leaking
        # into rendered HTML. File fields should only be in the files dict.
        # We check the form's base_fields to identify which fields are FileFields.
        file_field_names = {
            name
            for name, field in self.form_class.base_fields.items()
            if isinstance(field, FileField)
        }
        # TODO: better include this directly into ._data()?
        data = {
            key: value
            for key, value in self._data().items()
            if key not in file_field_names
        }
        kwargs = {
            # "initial": self.get_form_initial(), # TODO?
            # "prefix": self.get_prefix(), # TODO?
            "data": data,
            "files": self._temp_files,
        }
        return kwargs

    # def get_form_initial(self):
    #     return {}

    def get_form(self):
        """Returns a new form instance, initialized with data from the component
        attributes."""
        cls = self.form_class
        form = cls(**self.get_form_kwargs())
        self._add_alpine_models_to_fields(form)
        return form

    def _add_alpine_models_to_fields(self, form, prefix: str = None) -> None:
        """Connects the form's fields to the Tetra backend using x-model attributes."""
        for field_name, field in form.fields.items():
            if isinstance(field, FileField):
                # TODO: this is not necessary any more
                form.fields[field_name].widget.attrs.update(
                    {"@change": f"{field_name}=$event.target.files?.[0];"}
                )

                # form.fields[field_name].initial = getattr(self, field_name)
            else:
                if prefix is None:
                    prefix = ""
                if prefix and prefix[:-1] != ".":
                    prefix += "."
                # form.fields[field_name].initial = getattr(self, field_name)
                form.fields[field_name].widget.attrs.update(
                    {"x-model": prefix + field_name}
                )

    @public
    def validate(self, field_names: str | list = None) -> dict | None:
        """Validates the data using the form defined in `form_class`, and re-renders
        the component, without saving the form."""

        if self._validate_called:
            return self.form_errors
        self._validate_called = True

        if self._form:
            form_errors = self._form.errors.get_json_data(escape_html=True)

            # Synchronize existing errors with current form errors
            if self.form_errors:
                self.form_errors = {
                    key: value
                    for key, value in self.form_errors.items()
                    if key in form_errors
                }

            if type(field_names) is str:
                field_names = [field_names]
            # Update errors based on provided field names or include all form errors
            if field_names is not None:
                self.form_errors.update(
                    {
                        key: value
                        for key, value in form_errors.items()
                        if key in field_names
                    }
                )
            else:
                self.form_errors.update(form_errors)

    def form_valid(self, form: BaseForm) -> None:
        """This method is called when the form data is valid.
        It should save the form data.

        Parameters:
            form: The form instance that contains the validated form data.
        """

    def form_invalid(self, form: BaseForm) -> None:
        """Hook that gets called when form was validated with errors. Override this
        method to customize behaviour.

        Parameters:
            form: The form instance that contains the invalid form data.
        """

    def _call_public_method(
        self, request, method_name, children_state, *args
    ) -> JsonResponse | FileResponse:
        """Override to automatically submit form when files are uploaded.

        When files are uploaded via any component method call, they need to be
        processed immediately since browser file inputs are cleared on subsequent
        requests. This method checks if files were uploaded and automatically
        validates the form before calling the requested method.
        """
        # Check if any files were uploaded in this request
        has_uploaded_files = bool(self._temp_files)

        # If files were uploaded and the method is not already submit(),
        # automatically validate and process the form first
        if has_uploaded_files and method_name != "submit":
            self._process_form()

        # Now call the original method
        return super()._call_public_method(request, method_name, children_state, *args)

    def _process_form(self) -> None:
        """Internal method to validate and process the form.

        This is called both by submit() and automatically when files are uploaded.
        """
        self.form_submitted = True

        if self._form.is_valid():
            self.form_valid(self._form)
        else:
            self.form_errors = self._form.errors.get_json_data(escape_html=True)
            # set the possibly cleaned values back to the component's attributes
            for attr, value in self._form.cleaned_data.items():
                setattr(self, attr, value)
            self.form_invalid(self._form)

    @public
    def submit(self) -> None:
        """Submits the form.

        The component will validate the data against the form, and if the form is valid,
        it will call form_valid(), else form_invalid().
        """
        self._process_form()

    def _reset(self):
        """Internally resets all form fields to their defaults set in load(). This
        internal method is not exposed as a
        public API and can be called by load() too. For client side calls use reset() instead.
        """

        # first, clear all temporary files saved in the component
        try:
            for attr_name, file in self._temp_files.items():
                self._temp_files.pop(attr_name)
                self.client._setValueByName(attr_name, "")
                os.remove(file.name)
        except FileNotFoundError:
            # ignore any errors during file removal, file might be already deleted.
            pass

        # second, get a new form without any data, with just initial values of the
        # Form class

        self.form_submitted = False
        self._form = self.get_form()

        self._set_attrs_from_form_class_initial()

        # third, call load() again to apply the initial values meant by the component
        self._recall_load()

    def _set_attrs_from_form_class_initial(self):
        """Get 'initial' values from the form class and set them to the component's
        attributes."""
        # for field_name, field in self.form_class.base_fields.items():
        #     setattr(self, field_name, field.initial)
        #     if self.client and isinstance(field, FileField):
        #         # we additionally have to set the initial value of FileFields to an empty string
        #         # as the browser doesn't set the input field
        #         self.client._setValueByName(field_name, "")

    @public
    def reset(self):
        """Convenience method to be called by the frontend to reset the form.

        All values will be reset to the initial object values.
        """
        self._reset()


class ModelFormComponentMetaClass(FormComponentMetaClass):
    def __new__(cls, name, bases, dct):
        """additionally to FormComponentMetaClass.__new__(), add the object's pk to
        the component's attributes"""
        dct.setdefault("__annotations__", {})
        if "__abstract__" not in dct:
            model = dct.get("model", None)
            abstract = dct.get("__abstract__", False)
            # if no form_class nor model/fields is defined in the new class -> Error
            if (
                not dct.get("form_class", None)
                and (not model or not dct.get("fields", None))
                and not abstract
            ):
                raise AttributeError(
                    f"{name} class requires either a 'form_class' or 'model' and "
                    "'fields' attributes"
                )

            # if form_class is not provided, but model/fields, create a form_class on
            # the fly.
            if "form_class" in dct:
                form_class = dct["form_class"]
                if not issubclass(form_class, forms.ModelForm):
                    raise TypeError(
                        f"'{name}.form_class' must be a "
                        f"subclass of 'ModelForm', not '{form_class.__name__}'"
                    )
                # form_class' model must match the model's class
                elif model and model != form_class.Meta.model:
                    raise TypeError(
                        f"'{name}.form_class.model' must match "
                        f"'{name}.model' class, if provided. "
                        f"({form_class.Meta.model.__name__} != {model.__name__})"
                    )
            else:
                # logger.debug(
                #     f"Creating component attributes from model/fields for class {name}"
                # )
                form_class = modelform_factory(model=dct["model"], fields=dct["fields"])
                dct["form_class"] = form_class

            # So the new class *in any case* has a form_class
        return super().__new__(cls, name, bases, dct)


class ModelFormComponent(FormComponent, metaclass=ModelFormComponentMetaClass):
    """
    Component that can render a Model object using a ModelForm, load and
    save the given object, and validate it. This is basically the equivalent of the
    UpdateView/CreateView class for views in Django.

    Like FormComponent, the class mirrors all FormFields to the component's
    attributes which are then automatically available in the template and the client
    as Javascript/Alpine.js data. When a component attribute is set,
    it is automatically copied into the form instance's properties.

    Attributes:
        form_class (type(ModelForm)): the ModelForm class to use
        model (type(models.Model)): the Model class to use, as alternative to form_class
        fields (list[str]|str): the fields to use in the form, as list, or "__all__" to use all fields
        _context_object_name (str): the name of the context object in the template (TODO)
    """

    __abstract__ = True
    form_class: Optional[type[forms.ModelForm]] = None
    model: ModelBase = None
    object: models.Model = None
    fields: Optional[list[str]] = None
    _context_object_name = None  # TODO
    _excluded_props_from_saved_state = (
        FormComponent._excluded_props_from_saved_state
        + ["_context_object_name", "model", "fields"]
    )

    # def __init__(self, *args, **kwargs):
    #     if not self.model:
    #         self.model = self.get_form_class().Meta.model
    #     super().__init__(*args, **kwargs)

    def get_model(self):
        """Returns the model to use in this component."""
        if not self.model:
            self.model = self.get_form_class()._meta.model
        return self.model

    def get_form_class(self) -> type[forms.ModelForm]:
        """Returns the form class to use in this component."""

        # if there is an explicit form_class defined, use it
        if self.form_class:
            return self.form_class
        else:
            raise ImproperlyConfigured(
                f"'{self.__class__.__name__}' must either define 'form_class' "
                "or both 'model' and 'fields', or override 'get_form_class()'"
            )

    def _pre_load(self, object: models.Model = None, *args, **kwargs) -> None:
        """
        Automatically makes the object available in the component's attributes.

        Attributes:
            object (models.Model): the object to load into the component
        """
        if object:
            assert isinstance(object, self.get_model()), (
                self.__class__.__name__
                + " expects an instance of "
                + self.get_model().__name__
            )
        self.object = object or self.get_model()()
        self._copy_object_to_attrs()
        self._set_attrs_from_form_class_initial()

    def __setattr__(self, key, value):
        """Additionally to setting the component attribute, set it on the attached
        object, if applicable."""
        super().__setattr__(key, value)
        if self.object and key in {f.attname for f in self.object._meta.fields}:
            # if the key is a field of the 'self.object' model, set it on self.object
            setattr(self.object, key, value)

    def _reset(self):
        # set an empty model instance and initialize it with defaults
        self.object = self.get_model()()
        super()._reset()

    def get_form_kwargs(self):
        """Adds self.object as instance to the form kwargs."""
        kwargs = super().get_form_kwargs()
        kwargs["instance"] = self.object
        return kwargs

    def _copy_object_to_attrs(self):
        """Copies all object attributes to the component props."""
        for name in {f.attname for f in self.object._meta.fields}:
            # FIXME: this recursively sets self.object.<attname> again:
            setattr(self, name, getattr(self.object, name))

    def get_form(self, data=None, files=None, **kwargs):
        """Returns a form, bound to given model instance."""
        # create an instance of model() and copy component's attributes to it
        instance = self.object
        for field_name, field in self.get_form_class().base_fields.items():
            # print(f"set {field_name} to {getattr(self, field_name)}")
            if isinstance(field, ModelChoiceField):
                setattr(instance, f"{field_name}_id", getattr(self, field_name))
            else:
                setattr(instance, field_name, getattr(self, field_name))

        return super().get_form()

    # def _get_context_object_name(self, is_list=False):
    #     """
    #     Returns a descriptive name to use in the context in addition to the
    #     default 'object'/'object_list'.
    #     """
    #     if self._context_object_name is not None:
    #         return self._context_object_name
    #
    #     elif self.object is not None:
    #         return (
    #             f"{self.object.__name__.lower()}_list"
    #             if is_list
    #             else self.object.__name__.lower()
    #         )

    def form_valid(self, form: ModelForm) -> None:
        """Overrides the form_valid method to save the ModelForm data to the
        database.

        Override This method if you need custom functionality."""
        self.object = form.save(commit=True)
        # FIXME: reset is not always wanted, only in Create, not Update
        self._reset()


class DynamicFormMixin:
    """
    A mixin class that provides functionality for dynamically updating fields'
    attributes, based on other fields or circumstances.

    This class checks for the existence of special methods in your component,
    and calls them to get the current `queryset`, `choices`, `hidden`, `disabled`,
    and `required` status for the corresponding field.

    Usage:
        Create a method that is called whenever a parent field changes its value, using
        the `@public.watch("field_name")` decorator. The method itself could be
        empty. It is just needed as trigger to rerender the form.

        ```python


        @public.watch("make")
        def make_changed(self, value, old_value, attr) -> None:
            pass

        def get_engine_hidden(self) -> bool:
            # hide engine_type if make is Tesla
            return self.make == Make.objects.get(name="Tesla")
        ```

    The following methods can be optionally defined in your component, and will get
    called in time to determine the fields' attributes before rendering.

    Methods:
        get_<field_name>_queryset(): Returns the current queryset for the given field.
        get_<field_name>_disabled(): whether the given field should be
            disabled, depending on the values of other fields.
        get_<field_name>_hidden(): returns whether the given field should be
            hidden, depending on the values of other fields.
        get_<field_name>_required(): returns whether the given field is
            required, depending on the values of other fields.

        All these methods are normal instance methods and must return a boolean value.
    """

    def get_form(self, *args, **kwargs):
        """Updates dynamic fields when the form is created."""

        # get form and modify fields according to saved field state
        form = super().get_form(*args, **kwargs)
        # parents = self.field_dependencies.values()
        for field_name, field in form.fields.items():

            if update_method := getattr(self, f"get_{field_name}_disabled", None):
                form.fields[field_name].disabled = update_method()

            if update_method := getattr(self, f"get_{field_name}_hidden", None):
                form.fields[field_name].widget.attrs["hidden"] = update_method()

            if update_method := getattr(self, f"get_{field_name}_required", None):
                form.fields[field_name].required = update_method()

            # check if there is a dynamic queryset
            if update_method := getattr(self, f"get_{field_name}_queryset", None):
                form.fields[field_name].queryset = update_method()

            # check if there are dynamic choices
            if update_method := getattr(self, f"get_{field_name}_choices", None):
                form.fields[field_name].choices = update_method()
        return form


# this is just experimental
# class WizardFormComponent(FormComponent):
#     __abstract__ = True



import gzip
import base64
import logging
import pickle
from copy import copy
from typing import Any, TYPE_CHECKING
from io import BytesIO

from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from django.conf import settings
from django.db.models.query import QuerySet
from django.db.models import Model
from django.http import HttpRequest
from django.template import RequestContext, engines
from django.template.base import Origin
from django.template.loader_tags import BlockNode
from django.utils.functional import SimpleLazyObject, LazyObject

from .exceptions import ComponentError
from .utils import isclassmethod, NamedTemporaryUploadedFile
from .templates import InlineOrigin

if TYPE_CHECKING:
    from tetra.components import BasicComponent, Component


# Global flags to avoid circular imports
loading_libraries = False
find_libraries_done = False


class StateException(Exception):
    pass


logger = logging.getLogger(__name__)
picklers_by_type = {}
picklers_by_prefix = {}


class Pickler:
    """Base class for picklers.

    Define pickle and unpickle methods in a subclass and register it with
    the `register_pickler` decorator, e.g.

    ```python
    @register_pickler(MyObjectPickler, b"MyObject")
    ```
    """

    @staticmethod
    def pickle(obj: Any) -> bytes | None:
        pass

    @staticmethod
    def unpickle(bs: bytes) -> Any:
        pass


def register_pickler(obj_type, prefix: bytes):
    """Decorator for registering classes as picklers.

    See [Pickler](Pickler) class on how to implement a pickler.
    """

    def dec(cls):
        cls.obj_type = obj_type
        cls.prefix = prefix
        picklers_by_type[obj_type] = cls
        picklers_by_prefix[prefix] = cls
        return cls

    return dec


def resolve_lazy_object(lazy_object: Model | LazyObject) -> Model:
    """If it's a SimpleLazyObject, resolve it by accessing the underlying object."""
    if isinstance(lazy_object, SimpleLazyObject):
        return lazy_object._wrapped
    return lazy_object


# ----------- Picklers ------------
# see https://tetra.readthedocs.io/en/stable/state-security/


@register_pickler(QuerySet, b"QuerySet")
class PickleQuerySet(Pickler):
    @staticmethod
    def pickle(qs: QuerySet) -> bytes | None:
        return pickle.dumps(
            {
                "model": qs.model,
                "query": qs.query,
            }
        )

    @staticmethod
    def unpickle(bs: bytes) -> Any:
        data = pickle.loads(bs)
        qs = data["model"].objects.all()
        qs.query = data["query"]
        return qs


@register_pickler(Model, b"Model")
class PickleModel(Pickler):
    @staticmethod
    def pickle(obj: Model) -> bytes | None:
        return pickle.dumps(
            {
                "class": type(obj),
                "pk": obj.pk,
            }
        )

    @staticmethod
    def unpickle(bs: bytes) -> Model | None:
        data = pickle.loads(bs)
        model = data["class"]
        try:
            return model.objects.get(pk=data["pk"])
        except model.DoesNotExist:
            return None


@register_pickler(NamedTemporaryUploadedFile, b"NamedTemporaryUploadedFile")
class PickleNamedTemporaryUploadedFile(Pickler):
    @staticmethod
    def pickle(file: NamedTemporaryUploadedFile) -> bytes | None:
        # return a reference to file's temporary location
        return pickle.dumps(
            {
                "name": file.name,
                "size": file.size,
                "content_type": file.content_type,
                "temp_path": file.temporary_file_path(),
            }
        )

    @staticmethod
    def unpickle(bs: bytes) -> NamedTemporaryUploadedFile:
        data = pickle.loads(bs)
        return NamedTemporaryUploadedFile(
            name=data["name"],
            size=data["size"],
            content_type=data["content_type"],
            charset=settings.DEFAULT_CHARSET,
            temp_path=data["temp_path"],
        )


@register_pickler(BlockNode, b"BlockNode")
class PickleBlockNode(Pickler):
    def pickle(obj: Origin) -> bytes | None:
        origin = getattr(obj, "origin", None)
        if isinstance(origin, InlineOrigin) and hasattr(obj, "_path_key"):
            return pickle.dumps(
                {
                    "component": origin.component,
                    "block_path_key": obj._path_key,
                }
            )
        elif hasattr(obj, "_path_key"):
            return pickle.dumps(
                {
                    "loader_module": origin.loader.__class__.__module__,
                    "loader_name": origin.loader.__class__.__name__,
                    "template_name": origin.template_name,
                    "block_path_key": obj._path_key,
                }
            )
        return None

    def unpickle(bs: bytes) -> Origin:
        data = pickle.loads(bs)
        if "loader_name" in data:
            loader = engines["django"].engine.find_template_loader(
                f"{data['loader_module']}.{data['loader_name']}"
            )
            template = loader.get_template(data["template_name"])
            block = template.blocks_by_key[data["block_path_key"]]
            return block
        elif "component" in data:
            return data["component"]._template.blocks_by_key[data["block_path_key"]]
        else:
            raise TypeError("Unpicked data for template block incorrect.")


skip_check = {
    str,
    bytes,
    int,
    float,
    bool,
    dict,
    list,
    set,
    frozenset,
    tuple,
    range,
    bytes,
    bytearray,
    complex,
    type(None),
}


class StatePickler(pickle.Pickler):
    """Custom pickler for serializing component state with support for special object types.

    Extends pickle.Pickler to handle Django-specific objects and custom types that require
    special serialization logic. Uses registered picklers to serialize objects that cannot
    be pickled using standard pickle mechanisms.
    """

    def persistent_id(self, obj) -> bytes | None:
        """Generate a persistent ID for objects requiring custom serialization.

        This method is called by the pickle module for each object during serialization.
        It determines whether an object needs custom handling and returns a persistent ID
        that can be used to reconstruct the object during unpickling.

        The method performs the following steps:
        1. Skips basic types that can be pickled normally
        2. Handles Django Origin objects by removing unpickleable loader references
        3. Resolves lazy objects before pickling
        4. Finds and applies appropriate custom picklers for registered types

        Args:
            obj: The object to be pickled. Can be any Python object.

        Returns:
            bytes | None: A persistent ID in the format b"prefix:pickled_data" if a custom
                pickler is found and successfully serializes the object, None otherwise.
                When None is returned, the standard pickle mechanism is used.
        """
        if type(obj) in skip_check:
            return None

        # Template loaders are not pickleable, they are set as the 'loader' property
        # on block.origin which is an Origin obj. We set `obj.loader = None` to
        # stop it from erroring.
        # This will make error messages about templates a little less helpful, however
        # we have already rendered the template once and so it's not likely we will get
        # an exception.
        if isinstance(obj, Origin):
            obj.loader = None

        # for LazyObjects, resolve them before pickling
        obj = resolve_lazy_object(obj)

        pickler = None
        if type(obj) in picklers_by_type:
            pickler = picklers_by_type[type(obj)]
        else:
            for obj_type, pickler_option in picklers_by_type.items():
                if isinstance(obj, obj_type):
                    pickler = pickler_option
        if pickler:
            pickled = pickler.pickle(obj)
            if pickled is not None:
                return b":".join([pickler.prefix, pickled])

        return None


class StateUnpickler(pickle.Unpickler):
    def persistent_load(self, persistent_id: bytes | None) -> Any | None:
        prefix, data = persistent_id.split(b":", 1)
        if prefix in picklers_by_prefix:
            pickler = picklers_by_prefix[prefix]
            return pickler.unpickle(data)
        return None


def pickle_state(obj) -> bytes:
    out = BytesIO()
    try:
        StatePickler(out).dump(obj)
    except pickle.PicklingError as e:
        raise ComponentError(f"Failed to pickle state for {obj}: {e}")
    return out.getvalue()


def unpickle_state(data) -> Any | None:
    return StateUnpickler(BytesIO(data)).load()


def _get_fernet_for_request(request: HttpRequest) -> Fernet:
    if hasattr(request, "_tetra_state_fernet") and request._tetra_state_fernet:
        return request._tetra_state_fernet
    if not request.session.session_key:
        request.session.create()
    salt = "".join(
        [
            request.session.session_key,
            request.user.get_username() if request.user.is_authenticated else "",
        ]
    )
    hkdf = HKDF(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt.encode(),
        info=b"tetra-state",
    )
    key = base64.urlsafe_b64encode(hkdf.derive(settings.SECRET_KEY.encode()))
    fernet = Fernet(key)
    request._tetra_state_fernet = fernet
    return fernet


# These are the default context vars from BaseContext RequestContext, and few extra
keys_to_remove_from_context = [
    "True",
    "False",
    "None",
    "csrf_token",
    "request",
    "user",
    "perms",
    "messages",
    "DEFAULT_MESSAGE_LEVELS",
    "template",
    "_template",
]


def encode_component(component: "BasicComponent") -> str:
    """Serializes a component with its server state into a pickled state token."""
    fernet = _get_fernet_for_request(component.request)

    # TODO: this should be in component.__getstate__
    # We don't need a full context stack and don't want a RequestContext
    # TODO: in future would be better to patch RequestContext to keep track of what
    #  it adds that then they can be easily removed.
    original_context = component._context
    context = copy(original_context) or {}
    if isinstance(context, RequestContext):
        # Remove the top to layer of the context dicts:
        # 0: is the template defaults (True, False, None etc.)
        # 1: is the context added by RequestContext
        context.dicts = context.dicts[2:]
    if hasattr(context, "flatten"):
        context = context.flatten()
    for key in keys_to_remove_from_context:
        context.pop(key, None)
    for key in dir(component):
        # Remove vars from context that are filled from the component
        if not (
            key.startswith("_")
            or isclassmethod(getattr(component.__class__, key, None))
        ):
            context.pop(key, None)
    component._context = context
    # logger.debug(
    #     f"State before encoding: {component._data()}",
    # )
    pickled_state = pickle_state(component)
    component._context = original_context

    state_token = fernet.encrypt(gzip.compress(pickled_state)).decode()
    return state_token


def decode_component(state_token: str, request: HttpRequest) -> "Component":
    """Deserializes a pickled state token into a component, resuming its state."""
    fernet = _get_fernet_for_request(request)
    component: Component = unpickle_state(
        gzip.decompress(fernet.decrypt(state_token.encode()))
    )
    # logger.debug(
    #     f"State after decoding:  {component._data()}",
    # )
    return component



from django.template.base import Template, Origin
from django.template.loader_tags import BlockNode
from collections import defaultdict
import re


class InlineTemplate(Template):
    """Represents an "inline" template string within a component."""

    def get_exception_info(self, *args, **kwargs):
        ret = super().get_exception_info(*args, **kwargs)
        line_offset = self.origin.start_line - 1
        ret["top"] += line_offset
        ret["bottom"] += line_offset
        ret["line"] += line_offset
        ret["message"] = re.sub(
            r"(line )(\d+)(:)",
            lambda m: f"{m.group(1)}{int(m.group(2))+line_offset}{m.group(3)}",
            ret["message"],
        )
        ret["source_lines"] = [
            (line + line_offset, s) for (line, s) in ret["source_lines"]
        ]
        return ret


class InlineOrigin(Origin):
    def __init__(self, start_line=0, component=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_line = start_line
        self.component = component


def new_template_compile_nodelist(self):
    nodelist = Template._tetra_original_compile_nodelist(self)
    if not hasattr(self, "blocks_by_key"):
        self.blocks_by_key = {}
        annotate_nodelist(self, nodelist, [])
    return nodelist


new_template_compile_nodelist._tetra_patched = True  # type: ignore[attr-defined]


def monkey_patch_template():
    """
    Patch the `Template` class to replace its `compile_nodelist` method with a custom function.

    This function modifies the behavior of Django's `Template` class to use a new custom-defined
    `new_template_compile_nodelist` for compiling the nodelist. It ensures the patch is applied only
    once by checking the presence of a custom `_tetra_patched` attribute. Additionally, it preserves
    the original `compile_nodelist` method by storing it in an attribute
    `_tetra_original_compile_nodelist` if it hasn't been stored already.
    """
    if not getattr(Template.compile_nodelist, "_tetra_patched", False):
        if not hasattr(Template, "_tetra_original_compile_nodelist"):
            Template._tetra_original_compile_nodelist = Template.compile_nodelist
        Template.compile_nodelist = new_template_compile_nodelist


def annotate_nodelist(template, nodelist, path, memo=None):
    from .templatetags.tetra import ComponentNode

    if memo is None:
        memo = set()

    # Recursively annotate nodes, blocks, and component subtrees
    if nodelist and id(nodelist) not in memo:
        memo.add(id(nodelist))
        node_type_counter = defaultdict(int)
        for node in nodelist:
            if isinstance(node, BlockNode):
                node_key = f"block:{node.name}:{node_type_counter['block:'+node.name]}"
                node._path_key = "/".join([*path, node_key])
                template.blocks_by_key[node._path_key] = node
                annotate_nodelist(template, node.nodelist, [*path, node_key], memo)
                node_type_counter["block:" + node.name] += 1
            elif isinstance(node, ComponentNode):
                node_key = f"comp:{node.component_name}:{node_type_counter['comp:'+node.component_name]}"
                annotate_nodelist(template, node.nodelist, [*path, node_key], memo)
                node_type_counter["comp:" + node.component_name] += 1
            elif hasattr(node, "nodelist"):
                annotate_nodelist(template, node.nodelist, path, memo)
            elif hasattr(node, "branches"):
                # Handle nodes like IfNode that have multiple branches
                for branch in node.branches:
                    annotate_nodelist(template, branch[1], path, memo)
            elif hasattr(node, "nodelist_true"):
                # Handle IfNode in older Django versions or other nodes with nodelist_true/false
                annotate_nodelist(template, node.nodelist_true, path, memo)
                if hasattr(node, "nodelist_false"):
                    annotate_nodelist(template, node.nodelist_false, path, memo)



import json
import uuid
import warnings
from urllib.parse import urlsplit, urlunsplit

from django.contrib.messages import get_messages
from django.contrib.messages.storage.base import Message
from django.http import HttpRequest, QueryDict, FileResponse
from django.middleware.csrf import get_token
from django.utils.functional import cached_property

from .utils import (
    render_scripts,
    render_styles,
    TetraJSONEncoder,
    check_websocket_support,
)
from asgiref.sync import iscoroutinefunction, markcoroutinefunction, sync_to_async


class TetraMiddlewareException(Exception):
    pass


class TetraHttpRequest(HttpRequest):
    """This is just a dummy HttpRequest subclass that has a tetra attribute.

    This mainly exists so that the `request` attribute of components contains a
    `tetra` attribute, so IDEs can improve autocompletion. In fact, the `request`
    attribute is a WsgiRequest or `AsgiRequest` object.
    """

    tetra: "TetraDetails"


def inline_script_tag(source) -> str:
    return f"<script>{source}</script>"


def inline_styles_tag(source) -> str:
    return f"<styles>{source}</styles>"


class TetraDetails:
    _url: str = ""
    _websockets_available: bool = False

    # Based on HtmxDetails from htmx-django by adamchainz
    def __init__(self, request: HttpRequest) -> None:
        self.request = request
        self._url = self._get_header_value("T-Current-URL")

    def _get_header_value(self, name: str) -> str | None:
        value = self.request.headers.get(name) or None
        # if value:
        #     if self.request.headers.get(f"{name}-URI-AutoEncoded") == "true":
        #         value = unquote(value)
        return value

    def __bool__(self) -> bool:
        """returns True if 'T-Request' header is present and true, else False."""
        return self._get_header_value("T-Request") == "true"

    @cached_property
    def current_url(self) -> str | None:
        return self._url

    @cached_property
    def websockets_available(self) -> bool:
        return self._websockets_available

    # @cached_property
    # def new_url(self) -> str | None:
    #     return self._new_url if self._new_url else self.current_url

    @cached_property
    def current_url_abs_path(self) -> str | None:
        warnings.warn(
            "TetraDetails.current_url_abs_path is deprecated. Use TetraDetails.current_url_full_path instead.",
            DeprecationWarning,
        )
        return self.current_url_full_path

    @cached_property
    def current_url_full_path(self) -> str | None:
        """Returns the full path (including params) of the current URL in the
        browser.

        Example:
            When the browser URL is "https://example.com/foo/bar?baz=qux",
            current_url_path returns "/foo/bar?baz=qux".
        """
        url = self._url
        if url is not None:
            split = urlsplit(url)
            if (
                split.scheme == self.request.scheme
                and split.netloc == self.request.get_host()
            ):
                url = str(urlunsplit(split._replace(scheme="", netloc="")))

            else:
                url = None
        return url

    @cached_property
    def current_url_path(self) -> str:
        """Returns the path part of the current URL in the browser, without params.

        Example:
            When the browser URL is "https://example.com/foo/bar?baz=qux",
            current_url_path returns "/foo/bar".
        """
        value = self._url
        if value is not None:
            # get the path component from the URL
            return urlsplit(value).path or ""
        return ""

    @cached_property
    def url_query_params(self) -> QueryDict:
        """Returns the query parameters of the current URL in the browser, as a QueryDict.

        Example:
            When the browser URL is "https://example.com/foo/bar?baz=qux",
            url_query_params returns QueryDict({'baz': ['qux']}), so you can easily
            access the query parameters using `url_query_params['baz']`.
        """
        split = urlsplit(self._url)
        if (
            split.scheme == self.request.scheme
            and split.netloc == self.request.get_host()
        ):
            return QueryDict(split.query)
        return QueryDict()

    def set_url_path(self, path: str) -> None:
        """Replace path part of new URL with given path."""
        split = urlsplit(self._url)
        if (
            split.scheme == self.request.scheme
            and split.netloc == self.request.get_host()
        ):
            self._url = str(urlunsplit(split._replace(path=path)))

    def set_url(self, url: str) -> None:
        """Set new internal URL.

        This is needed if the browser URL will change during the request, and this
        should be reflected in the request.tetra object.
        """
        self._url = url

    def set_url_query_param(self, param, value):
        """Set/replace a query parameter of the current URL in the browser,
        as a QueryDict.
        """
        split = urlsplit(self._url)
        if (
            split.scheme == self.request.scheme
            and split.netloc == self.request.get_host()
        ):
            query_dict = QueryDict(split.query, mutable=True)
            query_dict[param] = value
            # replace the params in the split url
            self._url = str(urlunsplit(split._replace(query=query_dict.urlencode())))


class TetraMiddleware:
    async_capable = True
    sync_capable = True

    def __init__(self, get_response):
        self.get_response = get_response
        self._websocket_available = check_websocket_support()
        # Determine if we need async handling
        self._is_async = iscoroutinefunction(self.get_response)
        if self._is_async:
            markcoroutinefunction(self)

    def __call__(self, request):
        if self._is_async:
            return self.__acall__(request)
        else:
            return self._sync_call(request)

    async def __acall__(self, request):
        # Lightweight initialization - only create TetraDetails
        request.tetra = TetraDetails(request)
        request.tetra._websockets_available = self._websocket_available

        response = await self.get_response(request)

        # Only do expensive processing if Tetra components were actually used
        if hasattr(request, "tetra_components_used") and request.tetra_components_used:
            csrf_token = get_token(request)
            return self._process_response(request, response, csrf_token)

        # Fast path for non-Tetra requests - only handle messages
        return self._process_messages_only(request, response)

    def _sync_call(self, request):
        # Lightweight initialization - only create TetraDetails
        request.tetra = TetraDetails(request)
        request.tetra._websockets_available = self._websocket_available

        response = self.get_response(request)

        # Only do expensive processing if Tetra components were actually used
        if hasattr(request, "tetra_components_used") and request.tetra_components_used:
            csrf_token = get_token(request)
            return self._process_response(request, response, csrf_token)

        # Fast path for non-Tetra requests - only handle messages
        return self._process_messages_only(request, response)

    def _process_messages_only(self, request, response):
        """Fast path for non-Tetra requests - only process Django messages."""
        if isinstance(response, FileResponse):
            return response

        messages: list[Message] = []
        for message in get_messages(request):
            if not hasattr(message, "uid"):
                message.uid = str(uuid.uuid4())
            messages.append(message)

        if messages:
            response.headers["T-Messages"] = json.dumps(messages, cls=TetraJSONEncoder)

        return response

    def _process_response(self, request, response, csrf_token):
        """Full processing for Tetra requests - includes script/style injection."""
        messages: list[Message] = []

        if isinstance(response, FileResponse):
            return response

        for message in get_messages(request):
            # FIXME: maybe use a more efficient data structure for large number of
            #  messages

            # monkey patch an uid into the Message class to ensure it has a unique ID
            # if it doesn't have one already
            if not hasattr(message, "uid"):
                message.uid = str(uuid.uuid4())
            messages.append(message)
        if messages:
            # Put the messages list into the T-Messages header
            response.headers["T-Messages"] = json.dumps(messages, cls=TetraJSONEncoder)

        if (
            "Content-Type" not in response.headers
            or "text/html" not in response.headers["Content-Type"]
        ):
            return response
        if int(response.status_code) >= 500:
            return response

        if hasattr(request, "tetra_components_used") and request.tetra_components_used:
            if not hasattr(request, "tetra_scripts_placeholder_string"):
                raise TetraMiddlewareException(
                    "The {% tetra_scripts %} tag is required to be placed in the "
                    "page's <head> tag when using Tetra components."
                )
            if not hasattr(request, "tetra_styles_placeholder_string"):
                raise TetraMiddlewareException(
                    "The {% tetra_styles %} tag is required to be placed in the page's "
                    "<head> tag when using Tetra components."
                )
            if request.tetra_scripts_placeholder_string not in response.content:
                raise TetraMiddlewareException(
                    "Placeholder from {% tetra_scripts %} not found."
                )
            if request.tetra_styles_placeholder_string not in response.content:
                raise TetraMiddlewareException(
                    "Placeholder from {% tetra_styles %} not found."
                )

            content = response.content
            content = content.replace(
                request.tetra_scripts_placeholder_string,
                render_scripts(request, csrf_token).encode(),
            )
            content = content.replace(
                request.tetra_styles_placeholder_string, render_styles(request).encode()
            )
            response.content = content

        return response



import json
import logging

from django.http import HttpResponseNotFound, HttpResponseBadRequest, HttpResponse
from django.views.decorators.csrf import csrf_exempt, csrf_protect

from . import Library
from .utils import from_json, NamedTemporaryFileUploadHandler


logger = logging.getLogger(__name__)


@csrf_exempt
def component_method(request, *args, **kwargs):
    """Override default upload handlers, to create a "persistent" temporary file for
    file uploads that are done using Tetra methods."""
    request.upload_handlers = [NamedTemporaryFileUploadHandler(request)]
    return _component_method(request, *args, **kwargs)


@csrf_protect
def _component_method(
    request, app_name, library_name, component_name, method_name
) -> HttpResponse:
    if not request.method == "POST":
        return HttpResponseBadRequest()
    try:
        Component = Library.registry[app_name][library_name].components[component_name]
    except KeyError:
        return HttpResponseNotFound()

    if method_name not in (m["name"] for m in Component._public_methods):
        logger.warning(
            f"Tetra method was requested, but not found: {component_name}.{method_name}()"
        )
        return HttpResponseNotFound()

    try:
        # check if request includes multipart/form-data files
        if request.content_type == "multipart/form-data":
            component_state = from_json(request.POST["component_state"])
            # get files too and add them to the component state
            for key in request.FILES:
                component_state["data"][key] = request.FILES[key]

        # if the request is application-data/json, we need to decode it ourselves
        elif request.content_type == "application/json" and request.body:
            component_state = from_json(request.body.decode())
        else:
            logger.error("Unsupported content type: %s", request.content_type)
            return HttpResponseBadRequest()

    except json.decoder.JSONDecodeError as e:
        logger.error(e)
        return HttpResponseBadRequest()

    if not (
        isinstance(component_state, dict)
        and "args" in component_state
        and isinstance(component_state["args"], list)
    ):
        raise TypeError("Invalid component state args.")

    if not hasattr(request, "tetra_components_used"):
        request.tetra_components_used = set()
    request.tetra_components_used.add(Component)

    component = Component.from_state(component_state, request)

    logger.debug(
        f"Calling component method {component.__class__.__name__}.{method_name}()"
    )
    return component._call_public_method(
        request, method_name, component_state["children"], *component_state["args"]
    )



from typing import Literal

from channels.layers import get_channel_layer


class ComponentDispatcher:
    """
    A class to dispatch notifications, updates, data and removal requests of
    components to WebSocket connections.

    It provides convenient methods to asynchronously send data to client components.
    Beware that any of the methods are asynchronous and should be called inside an
    asynchronous context. If you use them from synchronous code, you will need to use
    `async_to_sync` from `asgiref.sync` to convert synchronous code to asynchronous:

    ```python
    from tetra.dispatcher import ComponentDispatcher

    await ComponentDispatcher.update_data(
    # or async_to_sync(ComponentDispatcher.update_data)(
        "notifications.news.headline",
        {
            "headline": "Breaking News!",
        }
    )
    ```
    """

    @staticmethod
    async def notify(group: str, event_name: str, data: dict) -> None:
        """
        Sends a JavaScript notification to all WebSocket connections in a group.

        Args:
           group: WebSocket group name to send the notification to.
           event_name: Name of the event to be sent, in the form "tetra:some-event"
           data: Dictionary containing the details of the event
        """
        channel_layer = get_channel_layer()
        await channel_layer.group_send(
            group,
            {
                "type": "notify",
                "event_name": event_name,
                "data": data,
            },
        )

    @staticmethod
    async def update_data(group: str, data: dict | None = None) -> None:
        """
        Sends data updates to public properties of all components that are subscribed to
        WebSocket connections in a group.

        Args:
           group: WebSocket group name to send the update to. This can be a normal
                group name or the predefined ones:
                * `user.{user_id}`
                * `session.{session_key}`
                * `broadcast`
           data: Dictionary containing the data to be sent to group members.
                IMPORTANT: All keys in this dict must match the components'
                public_properties names.
        """
        channel_layer = get_channel_layer()
        await channel_layer.group_send(
            group,
            {
                "type": "component.update_data",
                "group": group,
                "data": data or {},
            },
        )

    @staticmethod
    async def component_remove(group: str, component_id: str) -> None:
        """
        Send component removal notification to all WebSocket connections in a group.

        Args:
            group: WebSocket group name to send the removal notification to
            component_id: Unique identifier of the component to be removed
        """
        channel_layer = get_channel_layer()
        await channel_layer.group_send(
            group,
            {
                "type": "component.remove",
                "group": group,
                "component_id": component_id,
            },
        )

    @staticmethod
    async def subscription_response(
        group: str,
        status: Literal["subscribed", "unsubscribed", "resubscribed", "error"],
        message: str = "",
    ) -> None:
        """
        Send subscription confirmation to a WebSocket group.

        Args:
            group: WebSocket group name
            status: "subscribed", "unsubscribed", "resubscribed", or "error"
            message: Optional additional details
        """
        assert status in ["subscribed", "unsubscribed", "resubscribed", "error"]
        channel_layer = get_channel_layer()
        await channel_layer.group_send(
            group,
            {
                "type": "subscription.response",
                "group": group,
                "status": status,  # "subscribed", "unsubscribed", "error"
                "message": message or "",
            },
        )



import re
from typing import Optional

from django.conf import settings
from sourcetypes import django_html
from tetra import Component, public


def ensure_trailing_slash(path):
    if settings.APPEND_SLASH:
        return path if path.endswith("/") else path + "/"
    return path


class Router(Component):
    """
    A component that manages navigation and dynamic component switching.
    """

    routes: dict[str, str] = {}  # Map of path patterns to component names.
    current_component: str = public("")
    current_path: str = public("")

    # language=html
    template: django_html = """
    <div
        @popstate.window="navigate(window.location.pathname, false)"
        @tetra:navigate.window="navigate($event.detail.path)"
    >
        {% if current_component %}
            {% component =current_component / %}
        {% else %}
            {% slot "default" %}{% endslot %}
        {% endif %}
    </div>
    """

    def load(self, *args, **kwargs):
        if not self.current_component:
            self.navigate(self.request.path, push=False)

    @public
    def navigate(self, path, push=True):
        if push:
            self.client._pushUrl(path)

        self.current_path = path
        component_name = self._match_route(path)
        self.current_component = component_name or ""

    def _match_route(self, path) -> str:
        path = ensure_trailing_slash(path)

        # Exact match
        if path in self.routes:
            return self.routes[path]

        # Regex match
        for pattern, component_name in self.routes.items():
            if re.match(f"^{pattern}$", path):
                return component_name

        return ""


class Link(Component):
    """
    A component for navigating between routes.
    """

    to: str = ""
    label: str = ""
    active_class: str = "active"

    # language=html
    template: django_html = """
    <a
        {% ... attrs %}
        href="{{ to }}"
        @click.prevent="click()"
        :class="{ '{{ active_class }}': window.location.pathname === '{{ to }}' }"
    >
        {% if label %}{{ label }}{% else %}{% slot "default" %}{% endslot %}{% endif %}
    </a>
    """

    def load(self, to="#", label="", active_class="active", *args, **kwargs):
        self.to = to
        self.label = label
        self.active_class = active_class

    @public(update=False)
    def click(self):
        self.client._dispatch("tetra:navigate", {"path": self.to})


class Redirect(Component):
    """
    A component that redirects to another path when rendered.
    """

    to: str = ""

    template: django_html = "<div></div>"

    def load(self, to: str, *args, **kwargs):
        self.to = to

    def render(self, *args, **kwargs):
        self.client._dispatch("tetra:navigate", {"path": self.to})
        return super().render(*args, **kwargs)
